from pathlib import Path
from typing import Literal

from confz import BaseConfig, EnvSource


class OperationMode(BaseConfig):  # type: ignore[misc]
    environment: Literal["testing", "development", "production"] = "development"
    """The service's operation mode."""

    CONFIG_SOURCES = EnvSource(allow=["ENVIRONMENT"])


class ServiceConfig(BaseConfig):  # type: ignore[misc]
    """Service configuration.

    Variables can be set through environment variables prefixed with `SERVICE_` or an
    `.env` file containing those variables.
    """

    config_dir: Path = Path("config")
    """Configuration directory"""
    web_port: int = 8001
    """Web server port"""

    CONFIG_SOURCES = EnvSource(allow_all=True, prefix="SERVICE_", file=".env")


class WebServerConfig(BaseConfig):  # type: ignore[misc]
    """The service's web server configuration."""

    generate_web_settings: bool = False
    """Should generate web_settings.json file"""

    CONFIG_SOURCES = EnvSource(allow=["GENERATE_WEB_SETTINGS"])



import asyncio
import inspect
import logging
import sys
from typing import Any, TypedDict

from pydase.utils.helpers import get_object_attr_from_path

if sys.version_info < (3, 11):
    from typing_extensions import NotRequired
else:
    from typing import NotRequired

import click
import socketio  # type: ignore[import-untyped]

import pydase.utils.serialization.deserializer
import pydase.utils.serialization.serializer
from pydase.data_service.data_service_observer import DataServiceObserver
from pydase.data_service.state_manager import StateManager
from pydase.server.web_server.api.v1 import endpoints
from pydase.utils.logging import SocketIOHandler
from pydase.utils.serialization.serializer import SerializedObject

logger = logging.getLogger(__name__)

# These functions can be monkey-patched by other libraries at runtime
dump = pydase.utils.serialization.serializer.dump
sio_client_manager = None


class UpdateDict(TypedDict):
    """
    A TypedDict subclass representing a dictionary used for updating attributes in a
    DataService.

    Attributes:
    ----------
    access_path : string
        The full access path of the attribute to be updated.
    value : SerializedObject
        The serialized new value to be assigned to the attribute.
    """

    access_path: str
    value: SerializedObject


class TriggerMethodDict(TypedDict):
    access_path: str
    args: NotRequired[SerializedObject]
    kwargs: NotRequired[SerializedObject]


class RunMethodDict(TypedDict):
    """
    A TypedDict subclass representing a dictionary used for running methods from the
    exposed DataService.

    Attributes:
        name:
            The name of the method to be run.
        parent_path:
            The access path for the parent object of the method to be run. This is used
            to construct the full access path for the method. For example, for an method
            with access path 'attr1.list_attr[0].method_name', 'attr1.list_attr[0]'
            would be the parent_path.
        kwargs:
            The arguments passed to the method.
    """

    name: str
    parent_path: str
    kwargs: dict[str, Any]


def setup_sio_server(
    observer: DataServiceObserver,
    enable_cors: bool,
    loop: asyncio.AbstractEventLoop,
) -> socketio.AsyncServer:
    """
    Sets up and configures a Socket.IO asynchronous server.

    Args:
        observer:
            The observer managing state updates and communication.
        enable_cors:
            Flag indicating whether CORS should be enabled for the server.
        loop:
            The event loop in which the server will run.

    Returns:
        The configured Socket.IO asynchronous server.
    """

    state_manager = observer.state_manager

    if enable_cors:
        sio = socketio.AsyncServer(
            async_mode="aiohttp",
            cors_allowed_origins="*",
            client_manager=sio_client_manager,
        )
    else:
        sio = socketio.AsyncServer(
            async_mode="aiohttp",
            client_manager=sio_client_manager,
        )

    setup_sio_events(sio, state_manager)
    setup_logging_handler(sio)

    # Add notification callback to observer
    def sio_callback(
        full_access_path: str, value: Any, cached_value_dict: SerializedObject
    ) -> None:
        if cached_value_dict != {}:

            async def notify() -> None:
                try:
                    await sio.emit(
                        "notify",
                        {
                            "data": {
                                "full_access_path": full_access_path,
                                "value": cached_value_dict,
                            }
                        },
                    )
                except Exception as e:
                    logger.warning("Failed to send notification: %s", e)

            loop.create_task(notify())

    observer.add_notification_callback(sio_callback)

    return sio


def setup_sio_events(sio: socketio.AsyncServer, state_manager: StateManager) -> None:  # noqa: C901
    @sio.event  # type: ignore
    async def connect(sid: str, environ: Any) -> None:
        client_id_header = environ.get("HTTP_X_CLIENT_ID", None)
        remote_username_header = environ.get("HTTP_REMOTE_USER", None)

        if remote_username_header is not None:
            log_id = f"user={click.style(remote_username_header, fg='cyan')}"
        elif client_id_header is not None:
            log_id = f"id={click.style(client_id_header, fg='cyan')}"
        else:
            log_id = f"sid={click.style(sid, fg='cyan')}"

        async with sio.session(sid) as session:
            session["client_id"] = log_id
            logger.info("Client [%s] connected", session["client_id"])

    @sio.event  # type: ignore
    async def disconnect(sid: str) -> None:
        async with sio.session(sid) as session:
            logger.info("Client [%s] disconnected", session["client_id"])

    @sio.event  # type: ignore
    async def service_serialization(sid: str) -> SerializedObject:
        async with sio.session(sid) as session:
            logger.info(
                "Client [%s] requested service serialization", session["client_id"]
            )
        return state_manager.cache_manager.cache

    @sio.event
    async def update_value(sid: str, data: UpdateDict) -> SerializedObject | None:
        async with sio.session(sid) as session:
            logger.info(
                "Client [%s] is updating the value of '%s'",
                session["client_id"],
                data["access_path"],
            )
        try:
            endpoints.update_value(state_manager=state_manager, data=data)
        except Exception as e:
            logger.exception(e)
            return dump(e)
        return None

    @sio.event
    async def get_value(sid: str, access_path: str) -> SerializedObject:
        async with sio.session(sid) as session:
            logger.info(
                "Client [%s] is getting the value of '%s'",
                session["client_id"],
                access_path,
            )
        try:
            return endpoints.get_value(
                state_manager=state_manager, access_path=access_path
            )
        except Exception as e:
            logger.exception(e)
            return dump(e)

    @sio.event
    async def trigger_method(sid: str, data: TriggerMethodDict) -> Any:
        async with sio.session(sid) as session:
            logger.info(
                "Client [%s] is triggering the method '%s'",
                session["client_id"],
                data["access_path"],
            )
        try:
            method = get_object_attr_from_path(
                state_manager.service, data["access_path"]
            )
            if inspect.iscoroutinefunction(method):
                return await endpoints.trigger_async_method(
                    state_manager=state_manager, data=data
                )
            return endpoints.trigger_method(state_manager=state_manager, data=data)
        except Exception as e:
            logger.exception(e)
            return dump(e)


def setup_logging_handler(sio: socketio.AsyncServer) -> None:
    logging.getLogger().addHandler(SocketIOHandler(sio))
    logging.getLogger("pydase").addHandler(SocketIOHandler(sio))



import asyncio
import logging
import os
import signal
import sys
import threading
from pathlib import Path
from types import FrameType
from typing import Any, Protocol, TypedDict

from pydase import DataService
from pydase.config import ServiceConfig
from pydase.data_service.data_service_observer import DataServiceObserver
from pydase.data_service.state_manager import StateManager
from pydase.server.web_server import WebServer
from pydase.task.autostart import autostart_service_tasks
from pydase.utils.helpers import current_event_loop_exists

HANDLED_SIGNALS = (
    signal.SIGINT,  # Unix signal 2. Sent by Ctrl+C.
    signal.SIGTERM,  # Unix signal 15. Sent by `kill <pid>`.
)
if sys.platform == "win32":  # pragma: py-not-win32
    HANDLED_SIGNALS += (signal.SIGBREAK,)  # Windows signal 21. Sent by Ctrl+Break.

logger = logging.getLogger(__name__)


class AdditionalServerProtocol(Protocol):
    """
    A Protocol that defines the interface for additional servers.

    This protocol sets the standard for how additional servers should be implemented
    to ensure compatibility with the main Server class. The protocol requires that
    any server implementing it should have an __init__ method for initialization and a
    serve method for starting the server.

    Args:
        data_service_observer:
            Observer for the DataService, handling state updates and communication to
            connected clients through injected callbacks. Can be utilized to access the
            service and state manager, and to add custom state-update callbacks.
        host:
            Hostname or IP address where the server is accessible. Commonly '0.0.0.0' to
            bind to all network interfaces.
        port:
            Port number on which the server listens. Typically in the range 1024-65535
            (non-standard ports).
        **kwargs:
            Any additional parameters required for initializing the server. These
            parameters are specific to the server's implementation.
    """

    def __init__(
        self,
        data_service_observer: DataServiceObserver,
        host: str,
        port: int,
        **kwargs: Any,
    ) -> None: ...

    async def serve(self) -> Any:
        """Starts the server. This method should be implemented as an asynchronous
        method, which means that it should be able to run concurrently with other tasks.
        """


class AdditionalServer(TypedDict):
    """A TypedDict that represents the configuration for an additional server to be run
    alongside the main server.
    """

    server: type[AdditionalServerProtocol]
    """Server adhering to the
    [`AdditionalServerProtocol`][pydase.server.server.AdditionalServerProtocol]."""
    port: int
    """Port on which the server should run."""
    kwargs: dict[str, Any]
    """Additional keyword arguments that will be passed to the server's constructor """


class Server:
    """
    The `Server` class provides a flexible server implementation for the `DataService`.

    Args:
        service: The DataService instance that this server will manage.
        host: The host address for the server. Defaults to `'0.0.0.0'`, which means all
            available network interfaces.
        web_port: The port number for the web server. If set to None, it will use the
            port defined in
            [`ServiceConfig().web_port`][pydase.config.ServiceConfig.web_port]. Defaults
            to None.
        enable_web: Whether to enable the web server.
        filename: Filename of the file managing the service state persistence.
        additional_servers: A list of additional servers to run alongside the main
            server.
            Here's an example of how you might define an additional server:

            ```python
            class MyCustomServer:
                def __init__(
                    self,
                    data_service_observer: DataServiceObserver,
                    host: str,
                    port: int,
                    **kwargs: Any,
                ) -> None:
                    self.observer = data_service_observer
                    self.state_manager = self.observer.state_manager
                    self.service = self.state_manager.service
                    self.port = port
                    self.host = host
                    # handle any additional arguments...

                async def serve(self):
                    # code to start the server...
            ```

            And here's how you might add it to the `additional_servers` list when
            creating a `Server` instance:

            ```python
            server = Server(
                service=my_data_service,
                additional_servers=[
                    {
                        "server": MyCustomServer,
                        "port": 12345,
                        "kwargs": {"some_arg": "some_value"}
                    }
                ],
            )
            server.run()
            ```
        autosave_interval: Interval in seconds between automatic state save events.
            If set to `None`, automatic saving is disabled. Defaults to 30 seconds.
        **kwargs: Additional keyword arguments.
    """

    def __init__(  # noqa: PLR0913
        self,
        service: DataService,
        host: str = "0.0.0.0",
        web_port: int | None = None,
        enable_web: bool = True,
        filename: str | Path | None = None,
        additional_servers: list[AdditionalServer] | None = None,
        autosave_interval: float = 30.0,
        **kwargs: Any,
    ) -> None:
        if additional_servers is None:
            additional_servers = []
        self._service = service
        self._host = host
        if web_port is None:
            self._web_port = ServiceConfig().web_port
        else:
            self._web_port = web_port
        self._enable_web = enable_web
        self._kwargs = kwargs
        self._additional_servers = additional_servers
        self.should_exit = False
        self.servers: dict[str, asyncio.Future[Any]] = {}
        self._state_manager = StateManager(
            service=self._service,
            filename=filename,
            autosave_interval=autosave_interval,
        )
        self._observer = DataServiceObserver(self._state_manager)
        self._state_manager.load_state()
        autostart_service_tasks(self._service)
        if not current_event_loop_exists():
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)
        else:
            self._loop = asyncio.get_event_loop()

    def run(self) -> None:
        """
        Initializes the asyncio event loop and starts the server.

        This method should be called to start the server after it's been instantiated.
        """
        self._loop.run_until_complete(self.serve())

    async def serve(self) -> None:
        process_id = os.getpid()

        logger.info("Started server process [%s]", process_id)

        await self.startup()
        if self.should_exit:
            return
        await self.main_loop()
        await self.shutdown()

        logger.info("Finished server process [%s]", process_id)

    async def startup(self) -> None:
        self._loop.set_exception_handler(self.custom_exception_handler)
        self.install_signal_handlers()

        for server in self._additional_servers:
            addin_server = server["server"](
                data_service_observer=self._observer,
                host=self._host,
                port=server["port"],
                **server["kwargs"],
            )

            server_name = (
                addin_server.__module__ + "." + addin_server.__class__.__name__
            )

            server_task = self._loop.create_task(addin_server.serve())
            server_task.add_done_callback(self._handle_server_shutdown)
            self.servers[server_name] = server_task
        if self._enable_web:
            self._web_server = WebServer(
                data_service_observer=self._observer,
                host=self._host,
                port=self._web_port,
                **self._kwargs,
            )
            server_task = self._loop.create_task(self._web_server.serve())

            server_task.add_done_callback(self._handle_server_shutdown)
            self.servers["web"] = server_task

        self._loop.create_task(self._state_manager.autosave())

    def _handle_server_shutdown(self, task: asyncio.Task[Any]) -> None:
        """Handle server shutdown. If the service should exit, do nothing. Else, make
        the service exit."""

        if self.should_exit:
            return

        try:
            task.result()
        except Exception:
            self.should_exit = True

    async def main_loop(self) -> None:
        while not self.should_exit:
            await asyncio.sleep(0.1)

    async def shutdown(self) -> None:
        logger.info("Shutting down")

        logger.info("Saving data to %s.", self._state_manager.filename)
        self._state_manager.save_state()

        logger.debug("Cancelling servers")
        await self.__cancel_servers()
        logger.debug("Cancelling tasks")
        await self.__cancel_tasks()

    async def __cancel_servers(self) -> None:
        for server_name, task in self.servers.items():
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                logger.debug("Cancelled '%s' server.", server_name)
            except Exception as e:
                logger.exception("Unexpected exception: %s", e)

    async def __cancel_tasks(self) -> None:
        for task in asyncio.all_tasks(self._loop):
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                logger.debug("Cancelled task '%s'.", task.get_coro())
            except Exception as e:
                logger.exception("Unexpected exception: %s", e)

    def install_signal_handlers(self) -> None:
        if threading.current_thread() is not threading.main_thread():
            # Signals can only be listened to from the main thread.
            return

        for sig in HANDLED_SIGNALS:
            signal.signal(sig, self.handle_exit)

    def handle_exit(self, sig: int = 0, frame: FrameType | None = None) -> None:
        if self.should_exit and sig == signal.SIGINT:
            logger.warning("Received signal '%s', forcing exit...", sig)
            os._exit(1)
        else:
            self.should_exit = True
            logger.warning(
                "Received signal '%s', exiting... (CTRL+C to force quit)", sig
            )

    def custom_exception_handler(
        self, loop: asyncio.AbstractEventLoop, context: dict[str, Any]
    ) -> None:
        # if any background task creates an unhandled exception, shut down the entire
        # loop. It's possible we don't want to do this, maybe make this optional in the
        # future
        loop.default_exception_handler(context)

        # here we exclude most kinds of exceptions from triggering this kind of shutdown
        exc = context.get("exception")
        if type(exc) not in [RuntimeError, KeyboardInterrupt, asyncio.CancelledError]:
            if self._enable_web:

                async def emit_exception() -> None:
                    try:
                        await self._web_server._sio.emit(
                            "exception",
                            {
                                "data": {
                                    "exception": str(exc),
                                    "type": exc.__class__.__name__,
                                }
                            },
                        )
                    except Exception as e:
                        logger.exception("Failed to send notification: %s", e)

                loop.create_task(emit_exception())
        else:
            self.handle_exit()



import base64
import io
import logging
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.request import urlopen

from pydase.data_service.data_service import DataService

if TYPE_CHECKING:
    from matplotlib.figure import Figure

logger = logging.getLogger(__name__)


class Image(DataService):
    def __init__(self) -> None:
        super().__init__()
        self._value: str = ""
        self._format: str = ""

    @property
    def value(self) -> str:
        return self._value

    @property
    def format(self) -> str:
        return self._format

    def load_from_path(self, path: Path | str) -> None:
        with open(path, "rb") as image_file:
            image_data = image_file.read()
        format_ = self._get_image_format_from_bytes(image_data)
        if format_ is None:
            logger.error("Unsupported image format. Skipping...")
            return
        value_ = base64.b64encode(image_data)
        self._load_from_base64(value_, format_)

    def load_from_matplotlib_figure(self, fig: "Figure", format_: str = "png") -> None:
        buffer = io.BytesIO()
        fig.savefig(buffer, format=format_)
        value_ = base64.b64encode(buffer.getvalue())
        self._load_from_base64(value_, format_)

    def load_from_url(self, url: str) -> None:
        with urlopen(url) as response:
            image_data = response.read()
        format_ = self._get_image_format_from_bytes(image_data)
        if format_ is None:
            logger.error("Unsupported image format. Skipping...")
            return
        value_ = base64.b64encode(image_data)
        self._load_from_base64(value_, format_)

    def load_from_base64(self, value_: bytes, format_: str | None = None) -> None:
        if format_ is None:
            format_ = self._get_image_format_from_bytes(base64.b64decode(value_))
            if format_ is None:
                logger.warning(
                    "Format of passed byte string could not be determined. Skipping..."
                )
                return
        self._load_from_base64(value_, format_)

    def _load_from_base64(self, value_: bytes, format_: str) -> None:
        value = value_.decode("utf-8")
        self._value = value
        self._format = format_

    def _get_image_format_from_bytes(self, value_: bytes) -> str | None:
        format_map = {
            b"\xff\xd8": "JPEG",
            b"\x89PNG": "PNG",
            b"GIF": "GIF",
            b"RIFF": "WEBP",
        }
        for signature, format_name in format_map.items():
            if value_.startswith(signature):
                return format_name
        return None



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Literal, TypedDict

if TYPE_CHECKING:
    import pydase.units as u

logger = logging.getLogger(__name__)


class SignatureDict(TypedDict):
    parameters: dict[str, dict[str, Any]]
    return_annotation: dict[str, Any]


class SerializedObjectBase(TypedDict):
    full_access_path: str
    doc: str | None
    readonly: bool


class SerializedInteger(SerializedObjectBase):
    value: int
    type: Literal["int"]


class SerializedFloat(SerializedObjectBase):
    value: float
    type: Literal["float"]


class SerializedQuantity(SerializedObjectBase):
    value: u.QuantityDict
    type: Literal["Quantity"]


class SerializedBool(SerializedObjectBase):
    value: bool
    type: Literal["bool"]


class SerializedString(SerializedObjectBase):
    value: str
    type: Literal["str"]


class SerializedDatetime(SerializedObjectBase):
    type: Literal["datetime"]
    value: str


class SerializedEnum(SerializedObjectBase):
    name: str
    value: str
    type: Literal["Enum", "ColouredEnum"]
    enum: dict[str, Any]


class SerializedList(SerializedObjectBase):
    value: list[SerializedObject]
    type: Literal["list"]


class SerializedDict(SerializedObjectBase):
    value: dict[str, SerializedObject]
    type: Literal["dict"]


class SerializedNoneType(SerializedObjectBase):
    value: None
    type: Literal["NoneType"]


class SerializedNoValue(SerializedObjectBase):
    value: None
    type: Literal["None"]


SerializedMethod = TypedDict(
    "SerializedMethod",
    {
        "full_access_path": str,
        "value": Literal["RUNNING"] | None,
        "type": Literal["method"],
        "doc": str | None,
        "readonly": bool,
        "async": bool,
        "signature": SignatureDict,
        "frontend_render": bool,
    },
)


class SerializedException(SerializedObjectBase):
    name: str
    value: str
    type: Literal["Exception"]


DataServiceTypes = Literal[
    "DataService", "Image", "NumberSlider", "DeviceConnection", "Task"
]


class SerializedDataService(SerializedObjectBase):
    name: str
    value: dict[str, SerializedObject]
    type: DataServiceTypes


SerializedObject = (
    SerializedBool
    | SerializedFloat
    | SerializedInteger
    | SerializedString
    | SerializedDatetime
    | SerializedList
    | SerializedDict
    | SerializedNoneType
    | SerializedMethod
    | SerializedException
    | SerializedDataService
    | SerializedEnum
    | SerializedQuantity
    | SerializedNoValue
)
"""
This type can be any of the following:

- SerializedBool
- SerializedFloat
- SerializedInteger
- SerializedString
- SerializedDatetime
- SerializedList
- SerializedDict
- SerializedNoneType
- SerializedMethod
- SerializedException
- SerializedDataService
- SerializedEnum
- SerializedQuantity
- SerializedNoValue
"""



from typing import Any

import pydase.data_service.data_service
import pydase.task.task
from pydase.task.task_status import TaskStatus
from pydase.utils.helpers import is_property_attribute


def autostart_service_tasks(
    service: pydase.data_service.data_service.DataService,
) -> None:
    """Starts the service tasks defined with the `autostart` keyword argument.

    This method goes through the attributes of the passed service and its nested
    [`DataService`][pydase.DataService] instances and calls the start method on
    autostart-tasks.
    """

    for attr in dir(service):
        if is_property_attribute(service, attr) or attr in {
            "_observers",
            "__dict__",
        }:  # prevent eval of property attrs and recursion
            continue

        val = getattr(service, attr)
        if isinstance(val, pydase.task.task.Task):
            if val.autostart and val.status == TaskStatus.NOT_RUNNING:
                val.start()
            else:
                continue
        else:
            autostart_nested_service_tasks(val)


def autostart_nested_service_tasks(
    service: pydase.data_service.data_service.DataService | list[Any] | dict[Any, Any],
) -> None:
    if isinstance(service, pydase.DataService):
        autostart_service_tasks(service)
    elif isinstance(service, list):
        for entry in service:
            autostart_nested_service_tasks(entry)
    elif isinstance(service, dict):
        for entry in service.values():
            autostart_nested_service_tasks(entry)



import enum
import logging
from datetime import datetime
from typing import TYPE_CHECKING, Any, NoReturn, cast

import pydase
import pydase.components
import pydase.units as u
from pydase.utils.helpers import (
    get_component_classes,
)
from pydase.utils.serialization.types import (
    SerializedDatetime,
    SerializedException,
    SerializedObject,
)

if TYPE_CHECKING:
    from collections.abc import Callable


logger = logging.getLogger(__name__)


class Deserializer:
    @classmethod
    def deserialize(cls, serialized_object: SerializedObject) -> Any:
        """Deserialize `serialized_object` (a `dict`) to a Python object."""
        type_handler: dict[str | None, None | Callable[..., Any]] = {
            None: None,
            "int": cls.deserialize_primitive,
            "float": cls.deserialize_primitive,
            "bool": cls.deserialize_primitive,
            "str": cls.deserialize_primitive,
            "NoneType": cls.deserialize_primitive,
            "Quantity": cls.deserialize_quantity,
            "Enum": cls.deserialize_enum,
            "ColouredEnum": lambda serialized_object: cls.deserialize_enum(
                serialized_object, enum_class=pydase.components.ColouredEnum
            ),
            "list": cls.deserialize_list,
            "dict": cls.deserialize_dict,
            "method": cls.deserialize_method,
            "Exception": cls.deserialize_exception,
            "datetime": cls.deserialize_datetime,
        }

        # First go through handled types (as ColouredEnum is also within the components)
        handler = type_handler.get(serialized_object["type"])
        if handler:
            return handler(serialized_object)

        # Custom types like Components or DataService classes
        service_base_class = cls.get_service_base_class(serialized_object["type"])
        if service_base_class:
            return cls.deserialize_data_service(serialized_object, service_base_class)

        return None

    @classmethod
    def deserialize_primitive(cls, serialized_object: SerializedObject) -> Any:
        if serialized_object["type"] == "float":
            return float(serialized_object["value"])
        return serialized_object["value"]

    @classmethod
    def deserialize_quantity(cls, serialized_object: SerializedObject) -> Any:
        return u.convert_to_quantity(serialized_object["value"])  # type: ignore

    @classmethod
    def deserialize_datetime(cls, serialized_object: SerializedDatetime) -> datetime:
        return datetime.fromisoformat(serialized_object["value"])

    @classmethod
    def deserialize_enum(
        cls,
        serialized_object: SerializedObject,
        enum_class: type[enum.Enum] = enum.Enum,
    ) -> Any:
        return enum_class(serialized_object["name"], serialized_object["enum"])[  # type: ignore
            serialized_object["value"]
        ]

    @classmethod
    def deserialize_list(cls, serialized_object: SerializedObject) -> Any:
        return [
            cls.deserialize(item)
            for item in cast(list[SerializedObject], serialized_object["value"])
        ]

    @classmethod
    def deserialize_dict(cls, serialized_object: SerializedObject) -> Any:
        return {
            key: cls.deserialize(value)
            for key, value in cast(
                dict[str, SerializedObject], serialized_object["value"]
            ).items()
        }

    @classmethod
    def deserialize_method(cls, serialized_object: SerializedObject) -> Any:
        return

    @classmethod
    def deserialize_exception(cls, serialized_object: SerializedException) -> NoReturn:
        import builtins

        try:
            exception = getattr(builtins, serialized_object["name"])
        except AttributeError:
            exception = type(serialized_object["name"], (Exception,), {})  # type: ignore
        raise exception(serialized_object["value"])

    @staticmethod
    def get_service_base_class(type_name: str | None) -> type | None:
        for component_class in get_component_classes():
            if type_name == component_class.__name__:
                return component_class
        if type_name in ("DataService", "Task"):
            import pydase

            return pydase.DataService
        return None

    @classmethod  # TODO: this shouldn't be a class method
    def create_attr_property(cls, serialized_attr: SerializedObject) -> property:
        attr_name = serialized_attr["full_access_path"].split(".")[-1]

        def get(self) -> Any:  # type: ignore
            return getattr(self, f"_{attr_name}")

        get.__doc__ = serialized_attr["doc"]

        def set(self, value: Any) -> None:  # type: ignore
            return setattr(self, f"_{attr_name}", value)

        if serialized_attr["readonly"]:
            return property(get)
        return property(get, set)

    @classmethod
    def deserialize_data_service(
        cls, serialized_object: SerializedObject, base_class: type
    ) -> Any:
        def create_proxy_class(serialized_object: SerializedObject) -> type:
            class_bases = (base_class,)
            class_attrs = {}

            # Process and add properties based on the serialized object
            for key, value in cast(
                dict[str, SerializedObject], serialized_object["value"]
            ).items():
                if value["type"] != "method":
                    class_attrs[key] = cls.create_attr_property(value)
                    # Initialize a placeholder for the attribute to avoid AttributeError
                    class_attrs[f"_{key}"] = cls.deserialize(value)

            # Create the dynamic class with the given name and attributes
            return type(serialized_object["name"], class_bases, class_attrs)  # type: ignore

        return create_proxy_class(serialized_object)()


def loads(serialized_object: SerializedObject) -> Any:
    """Deserialize `serialized_object` (a `dict`) to a Python object."""
    return Deserializer.deserialize(serialized_object)



import asyncio
import logging
from collections.abc import Callable
from copy import deepcopy
from typing import TYPE_CHECKING, cast

import socketio  # type: ignore

import pydase.components
from pydase.client.proxy_loader import ProxyClassMixin
from pydase.utils.helpers import get_attribute_doc
from pydase.utils.serialization.types import SerializedDataService, SerializedObject

logger = logging.getLogger(__name__)


class ProxyClass(ProxyClassMixin, pydase.components.DeviceConnection):
    """
    A proxy class that serves as the interface for interacting with device connections
    via a socket.io client in an asyncio environment.

    Args:
        sio_client:
            The socket.io client instance used for asynchronous communication with the
            pydase service server.
        loop:
            The event loop in which the client operations are managed and executed.
        reconnect:
            The method that is called periodically when the client is not connected.

    This class is used to create a proxy object that behaves like a local representation
    of a remote pydase service, facilitating direct interaction as if it were local
    while actually communicating over network protocols.
    It can also be used as an attribute of a pydase service itself, e.g.

    ```python
    import pydase


    class MyService(pydase.DataService):
        proxy = pydase.Client(
            hostname="...", port=8001, block_until_connected=False
        ).proxy


    if __name__ == "__main__":
        service = MyService()
        server = pydase.Server(service, web_port=8002).run()
    ```
    """

    def __init__(
        self,
        sio_client: socketio.AsyncClient,
        loop: asyncio.AbstractEventLoop,
        reconnect: Callable[..., None],
    ) -> None:
        if TYPE_CHECKING:
            self._service_representation: None | SerializedObject = None

        super().__init__()
        pydase.components.DeviceConnection.__init__(self)
        self._initialise(sio_client=sio_client, loop=loop)
        object.__setattr__(self, "_service_representation", None)
        self.reconnect = reconnect

    def serialize(self) -> SerializedObject:
        if self._service_representation is None:
            serialization_future = cast(
                asyncio.Future[SerializedDataService],
                asyncio.run_coroutine_threadsafe(
                    self._sio.call("service_serialization"), self._loop
                ),
            )
            # need to use object.__setattr__ to not trigger an observer notification
            object.__setattr__(
                self, "_service_representation", serialization_future.result()
            )
            if TYPE_CHECKING:
                self._service_representation = serialization_future.result()

        device_connection_value = cast(
            dict[str, SerializedObject],
            pydase.components.DeviceConnection().serialize()["value"],
        )

        readonly = False
        doc = get_attribute_doc(self)
        obj_name = self.__class__.__name__

        value = {
            **cast(
                dict[str, SerializedObject],
                # need to deepcopy to not overwrite the _service_representation dict
                # when adding a prefix with add_prefix_to_full_access_path
                deepcopy(self._service_representation["value"]),
            ),
            **device_connection_value,
        }

        return {
            "full_access_path": "",
            "name": obj_name,
            "type": "DeviceConnection",
            "value": value,
            "readonly": readonly,
            "doc": doc,
        }

    def connect(self) -> None:
        if not self._sio.reconnection or self._sio.reconnection_attempts > 0:
            self.reconnect(block_until_connected=False)



import inspect
import logging
import re
from collections.abc import Callable
from itertools import chain
from typing import Any

logger = logging.getLogger(__name__)


def parse_serialized_key(serialized_key: str) -> str | int | float:
    """
    Parse a serialized key and convert it to an appropriate type (int, float, or str).

    Args:
        serialized_key: str
            The serialized key, which might be enclosed in brackets and quotes.

    Returns:
        int | float | str:
            The processed key as an integer, float, or unquoted string.

    Examples:
        ```python
        print(parse_serialized_key("attr_name"))  # Outputs: attr_name  (str)
        print(parse_serialized_key("[123]"))  # Outputs: 123  (int)
        print(parse_serialized_key("[12.3]"))  # Outputs: 12.3  (float)
        print(parse_serialized_key("['hello']"))  # Outputs: hello  (str)
        print(parse_serialized_key('["12.34"]'))  # Outputs: 12.34  (str)
        print(parse_serialized_key('["complex"]'))  # Outputs: complex  (str)
        ```
    """

    # Strip outer brackets if present
    if serialized_key.startswith("[") and serialized_key.endswith("]"):
        serialized_key = serialized_key[1:-1]

    # Strip quotes if the resulting string is quoted
    if serialized_key.startswith(("'", '"')) and serialized_key.endswith(("'", '"')):
        return serialized_key[1:-1]

    # Try converting to float or int if the string is not quoted
    try:
        return float(serialized_key) if "." in serialized_key else int(serialized_key)
    except ValueError:
        # Return the original string if it's not a valid number
        return serialized_key


def parse_full_access_path(path: str) -> list[str]:
    """
    Splits a full access path into its atomic parts, separating attribute names, numeric
    indices (including floating points), and string keys within indices.

    Args:
        path: str
            The full access path string to be split into components.

    Returns:
        list[str]
            A list of components that make up the path, including attribute names,
            numeric indices, and string keys as separate elements.

    Example:
        >>> parse_full_access_path('dict_attr["some_key"].attr_name["other_key"]')
        ["dict_attr", '["some_key"]', "attr_name", '["other_key"]']
    """
    # Matches:
    # \w+ - Words
    # \[\d+\.\d+\] - Floating point numbers inside brackets
    # \[\d+\] - Integers inside brackets
    # \["[^"]*"\] - Double-quoted strings inside brackets
    # \['[^']*'\] - Single-quoted strings inside brackets
    pattern = r'\w+|\[\d+\.\d+\]|\[\d+\]|\["[^"]*"\]|\[\'[^\']*\']'
    return re.findall(pattern, path)


def get_path_from_path_parts(path_parts: list[str]) -> str:
    """Creates the full access path from its atomic parts.

    The reverse function is given by `parse_full_access_path`.

    Args:
        path_parts: list[str]
            A list of components that make up the path, including attribute names,
            numeric indices and string keys enclosed in square brackets as separate
            elements.
    Returns:
        str
            The full access path corresponding to the path_parts.
    """

    path = ""
    for path_part in path_parts:
        if not path_part.startswith("[") and path != "":
            path += "."
        path += path_part
    return path


def get_attribute_doc(attr: Any) -> str | None:
    """This function takes an input attribute attr and returns its documentation
    string if it's different from the documentation of its type, otherwise,
    it returns None.
    """
    attr_doc = inspect.getdoc(attr)
    attr_class_doc = inspect.getdoc(type(attr))
    return attr_doc if attr_class_doc != attr_doc else None


def get_class_and_instance_attributes(obj: object) -> dict[str, Any]:
    """Dictionary containing all attributes (both instance and class level) of a
    given object.

    If an attribute exists at both the instance and class level,the value from the
    instance attribute takes precedence.
    """

    return dict(chain(type(obj).__dict__.items(), obj.__dict__.items()))


def get_object_by_path_parts(target_obj: Any, path_parts: list[str]) -> Any:
    """Gets nested attribute of `target_object` specified by `path_parts`.

    Raises:
        AttributeError: Attribute does not exist.
        KeyError: Key in dict does not exist.
        IndexError: Index out of list range.
        TypeError: List index in the path is not a valid integer.
    """
    for part in path_parts:
        if part.startswith("["):
            deserialized_part = parse_serialized_key(part)
            target_obj = target_obj[deserialized_part]
        else:
            target_obj = getattr(target_obj, part)
    return target_obj


def get_object_attr_from_path(target_obj: Any, path: str) -> Any:
    """
    Traverse the object tree according to the given path.

    Args:
        target_obj: The root object to start the traversal from.
        path: Access path of the object.

    Returns:
        The attribute at the end of the path. If the path includes a list index,
        the function returns the specific item at that index. If an attribute in
        the path does not exist, the function logs a debug message and returns None.

    Raises:
        AttributeError: Attribute does not exist.
        KeyError: Key in dict does not exist.
        IndexError: Index out of list range.
        TypeError: List index in the path is not a valid integer.
    """
    path_parts = parse_full_access_path(path)
    return get_object_by_path_parts(target_obj, path_parts)


def get_task_class() -> type:
    from pydase.task.task import Task

    return Task


def get_component_classes() -> list[type]:
    """
    Returns references to the component classes in a list.
    """
    import pydase.components

    return [
        getattr(pydase.components, cls_name) for cls_name in pydase.components.__all__
    ]


def get_data_service_class_reference() -> Any:
    import pydase.data_service.data_service

    return getattr(pydase.data_service.data_service, "DataService")


def is_property_attribute(target_obj: Any, access_path: str) -> bool:
    path_parts = parse_full_access_path(access_path)
    target_obj = get_object_by_path_parts(target_obj, path_parts[:-1])

    # don't have to check if target_obj is dict or list as their content cannot be
    # properties -> always return False then
    return isinstance(getattr(type(target_obj), path_parts[-1], None), property)


def function_has_arguments(func: Callable[..., Any]) -> bool:
    sig = inspect.signature(func)
    parameters = dict(sig.parameters)
    # Remove 'self' parameter for instance methods.
    parameters.pop("self", None)

    # Check if there are any parameters left which would indicate additional arguments.
    return len(parameters) > 0


def is_descriptor(obj: object) -> bool:
    """Check if an object is a descriptor."""

    # Exclude functions, methods, builtins and properties
    if (
        inspect.isfunction(obj)
        or inspect.ismethod(obj)
        or inspect.isbuiltin(obj)
        or isinstance(obj, property)
    ):
        return False

    # Check if it has any descriptor methods
    return any(hasattr(obj, method) for method in ("__get__", "__set__", "__delete__"))


def current_event_loop_exists() -> bool:
    """Check if an event loop has been set."""
    import asyncio

    return asyncio.get_event_loop_policy()._local._loop is not None  # type: ignore



import inspect
from collections.abc import Callable
from typing import Any

from pydase.utils.helpers import function_has_arguments


class FunctionDefinitionError(Exception):
    pass


def frontend(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to mark a [`DataService`][pydase.DataService] method for frontend
    rendering. Ensures that the method does not contain arguments, as they are not
    supported for frontend rendering.
    """

    if function_has_arguments(func):
        raise FunctionDefinitionError(
            "The @frontend decorator requires functions without arguments. Function "
            f"'{func.__name__}' has at least one argument. "
            "Please remove the argument(s) from this function to use it with the "
            "@frontend decorator."
        )

    # Mark the function for frontend display.
    func._display_in_frontend = True  # type: ignore
    return func


def render_in_frontend(func: Callable[..., Any]) -> bool:
    """Determines if the method should be rendered in the frontend.

    It checks if the "@frontend" decorator was used or the method is a coroutine."""

    if inspect.iscoroutinefunction(func):
        return True

    try:
        return func._display_in_frontend  # type: ignore
    except AttributeError:
        return False



import asyncio
import logging
import sys
import threading
import urllib.parse
from types import TracebackType
from typing import TYPE_CHECKING, Any, TypedDict, cast

import socketio  # type: ignore

from pydase.client.proxy_class import ProxyClass
from pydase.client.proxy_loader import ProxyLoader
from pydase.utils.serialization.deserializer import loads
from pydase.utils.serialization.types import SerializedDataService, SerializedObject

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


logger = logging.getLogger(__name__)


class NotifyDataDict(TypedDict):
    full_access_path: str
    value: SerializedObject


class NotifyDict(TypedDict):
    data: NotifyDataDict


def asyncio_loop_thread(loop: asyncio.AbstractEventLoop) -> None:
    asyncio.set_event_loop(loop)
    loop.run_forever()


class Client:
    """
    A client for connecting to a remote pydase service using socket.io. This client
    handles asynchronous communication with a service, manages events such as
    connection, disconnection, and updates, and ensures that the proxy object is
    up-to-date with the server state.

    Args:
        url:
            The URL of the pydase Socket.IO server. This should always contain the
            protocol and the hostname.
        block_until_connected:
            If set to True, the constructor will block until the connection to the
            service has been established. This is useful for ensuring the client is
            ready to use immediately after instantiation. Default is True.
        sio_client_kwargs:
            Additional keyword arguments passed to the underlying
            [`AsyncClient`][socketio.AsyncClient]. This allows fine-tuning of the
            client's behaviour (e.g., reconnection attempts or reconnection delay).
            Default is an empty dictionary.
        client_id: Client identification that will be shown in the server logs this
            client is connecting to. This ID is passed as a `X-Client-Id` header in the
            HTTP(s) request. Defaults to None.

    Example:
        The following example demonstrates a `Client` instance that connects to another
        pydase service, while customising some of the connection settings for the
        underlying [`AsyncClient`][socketio.AsyncClient].

        ```python
        pydase.Client(url="ws://localhost:8001", sio_client_kwargs={
            "reconnection_attempts": 2,
            "reconnection_delay": 2,
            "reconnection_delay_max": 8,
        })
        ```

        When connecting to a server over a secure connection (i.e., the server is using
        SSL/TLS encryption), make sure that the `wss` protocol is used instead of `ws`:

        ```python
        pydase.Client(url="wss://my-service.example.com")
        ```
    """

    def __init__(
        self,
        *,
        url: str,
        block_until_connected: bool = True,
        sio_client_kwargs: dict[str, Any] = {},
        client_id: str | None = None,
    ):
        # Parse the URL to separate base URL and path prefix
        parsed_url = urllib.parse.urlparse(url)

        # Construct the base URL without the path
        self._base_url = urllib.parse.urlunparse(
            (parsed_url.scheme, parsed_url.netloc, "", "", "", "")
        )

        # Store the path prefix (e.g., "/service" in "ws://localhost:8081/service")
        self._path_prefix = parsed_url.path.rstrip("/")  # Remove trailing slash if any
        self._url = url
        self._sio = socketio.AsyncClient(**sio_client_kwargs)
        self._loop = asyncio.new_event_loop()
        self._client_id = client_id
        self.proxy = ProxyClass(
            sio_client=self._sio, loop=self._loop, reconnect=self.connect
        )
        """A proxy object representing the remote service, facilitating interaction as
        if it were local."""
        self._thread = threading.Thread(
            target=asyncio_loop_thread, args=(self._loop,), daemon=True
        )
        self._thread.start()
        self.connect(block_until_connected=block_until_connected)

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.disconnect()

    def connect(self, block_until_connected: bool = True) -> None:
        connection_future = asyncio.run_coroutine_threadsafe(
            self._connect(), self._loop
        )
        if block_until_connected:
            connection_future.result()

    def disconnect(self) -> None:
        connection_future = asyncio.run_coroutine_threadsafe(
            self._disconnect(), self._loop
        )
        connection_future.result()

    async def _connect(self) -> None:
        logger.debug("Connecting to server '%s' ...", self._url)
        await self._setup_events()

        headers = {}
        if self._client_id is not None:
            headers["X-Client-Id"] = self._client_id

        await self._sio.connect(
            url=self._base_url,
            headers=headers,
            socketio_path=f"{self._path_prefix}/ws/socket.io",
            transports=["websocket"],
            retry=True,
        )

    async def _disconnect(self) -> None:
        await self._sio.disconnect()

    async def _setup_events(self) -> None:
        self._sio.on("connect", self._handle_connect)
        self._sio.on("disconnect", self._handle_disconnect)
        self._sio.on("notify", self._handle_update)

    async def _handle_connect(self) -> None:
        logger.debug("Connected to '%s' ...", self._url)
        serialized_object = cast(
            SerializedDataService, await self._sio.call("service_serialization")
        )
        ProxyLoader.update_data_service_proxy(
            self.proxy, serialized_object=serialized_object
        )
        serialized_object["type"] = "DeviceConnection"
        if self.proxy._service_representation is not None:
            # need to use object.__setattr__ to not trigger an observer notification
            object.__setattr__(self.proxy, "_service_representation", serialized_object)

            if TYPE_CHECKING:
                self.proxy._service_representation = serialized_object  # type: ignore
        self.proxy._notify_changed("", self.proxy)
        self.proxy._connected = True

    async def _handle_disconnect(self) -> None:
        logger.debug("Disconnected from '%s' ...", self._url)
        self.proxy._connected = False

    async def _handle_update(self, data: NotifyDict) -> None:
        self.proxy._notify_changed(
            data["data"]["full_access_path"],
            loads(data["data"]["value"]),
        )



import inspect
import logging
import re
from typing import Any

from pydase.observer_pattern.observable.observable import Observable
from pydase.observer_pattern.observer.observer import Observer
from pydase.utils.helpers import is_descriptor

logger = logging.getLogger(__name__)


def reverse_dict(original_dict: dict[str, list[str]]) -> dict[str, list[str]]:
    reversed_dict: dict[str, list[str]] = {
        value: [] for values in original_dict.values() for value in values
    }
    for key, values in original_dict.items():
        for value in values:
            reversed_dict[value].append(key)
    return reversed_dict


def get_property_dependencies(prop: property, prefix: str = "") -> list[str]:
    source_code_string = inspect.getsource(prop.fget)  # type: ignore[arg-type]
    pattern = r"self\.([^\s\{\}]+)"
    matches = re.findall(pattern, source_code_string)
    return [prefix + match for match in matches if "(" not in match]


class PropertyObserver(Observer):
    def __init__(self, observable: Observable) -> None:
        super().__init__(observable)
        self._update_property_deps_dict()

    def _update_property_deps_dict(self) -> None:
        self.property_deps_dict = reverse_dict(
            self._get_properties_and_their_dependencies(self.observable)
        )

    def _get_properties_and_their_dependencies(
        self, obj: Observable, prefix: str = ""
    ) -> dict[str, list[str]]:
        deps: dict[str, Any] = {}

        self._process_observable_properties(obj, deps, prefix)
        self._process_nested_observables_properties(obj, deps, prefix)

        return deps

    def _process_observable_properties(
        self, obj: Observable, deps: dict[str, Any], prefix: str
    ) -> None:
        for k, value in inspect.getmembers(type(obj)):
            prefix = (
                f"{prefix}." if prefix != "" and not prefix.endswith(".") else prefix
            )
            key = f"{prefix}{k}"
            if isinstance(value, property):
                deps[key] = get_property_dependencies(value, prefix)

    def _process_nested_observables_properties(
        self, obj: Observable, deps: dict[str, Any], prefix: str
    ) -> None:
        for k, value in {**vars(type(obj)), **vars(obj)}.items():
            actual_value = value
            prefix = (
                f"{prefix}." if prefix != "" and not prefix.endswith(".") else prefix
            )
            parent_path = f"{prefix}{k}"

            # Get value from descriptor
            if not isinstance(value, property) and is_descriptor(value):
                actual_value = getattr(obj, k)

            if isinstance(actual_value, Observable):
                new_prefix = f"{parent_path}."
                deps.update(
                    self._get_properties_and_their_dependencies(
                        actual_value, new_prefix
                    )
                )
            elif isinstance(value, list | dict):
                self._process_collection_item_properties(
                    actual_value, deps, parent_path
                )

    def _process_collection_item_properties(
        self,
        collection: list[Any] | dict[str, Any],
        deps: dict[str, Any],
        parent_path: str,
    ) -> None:
        if isinstance(collection, list):
            for i, item in enumerate(collection):
                if isinstance(item, Observable):
                    new_prefix = f"{parent_path}[{i}]"
                    deps.update(
                        self._get_properties_and_their_dependencies(item, new_prefix)
                    )
        elif isinstance(collection, dict):
            for key, val in collection.items():
                if isinstance(val, Observable):
                    new_prefix = f'{parent_path}["{key}"]'
                    deps.update(
                        self._get_properties_and_their_dependencies(val, new_prefix)
                    )



import asyncio
import contextlib
import json
import logging
import os
from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

from pydase.data_service.data_service_cache import DataServiceCache
from pydase.utils.helpers import (
    get_object_by_path_parts,
    is_property_attribute,
    parse_full_access_path,
    parse_serialized_key,
)
from pydase.utils.serialization.deserializer import loads
from pydase.utils.serialization.serializer import (
    SerializationPathError,
    SerializedObject,
    generate_serialized_data_paths,
    get_nested_dict_by_path,
    serialized_dict_is_nested_object,
)

if TYPE_CHECKING:
    from pydase import DataService

logger = logging.getLogger(__name__)


def load_state(func: Callable[..., Any]) -> Callable[..., Any]:
    """This function should be used as a decorator on property setters to indicate that
    the value should be loaded from the JSON file.

    Example:
        ```python
        class Service(pydase.DataService):
            _name = "Service"

            @property
            def name(self) -> str:
                return self._name

            @name.setter
            @load_state
            def name(self, value: str) -> None:
                self._name = value
        ```
    """

    func._load_state = True  # type: ignore[attr-defined]
    return func


def has_load_state_decorator(prop: property) -> bool:
    """Determines if the property's setter method is decorated with the `@load_state`
    decorator.
    """

    try:
        return prop.fset._load_state  # type: ignore[union-attr]
    except AttributeError:
        return False


class StateManager:
    """
    Manages the state of a DataService instance, serving as both a cache and a
    persistence layer. It provides fast access to the most recently known state of the
    service and ensures consistent state updates across connected clients and service
    restarts.

    The StateManager is used by the web server to apply updates to service attributes
    and to serve the current state to newly connected clients. Internally, it creates a
    [`DataServiceCache`][pydase.data_service.data_service_cache.DataServiceCache]
    instance to track the state of public attributes and properties.

    The StateManager also handles state persistence: it can load a previously saved
    state from disk at startup and periodically autosave the current state to a file
    during runtime.

    Args:
        service: The DataService instance whose state is being managed.
        filename: The file name used for loading and storing the DataService's state.
            If provided, the state is loaded from this file at startup and saved to it
            on shutdown or at regular intervals.
        autosave_interval: Interval in seconds between automatic state save events.
            If set to `None`, automatic saving is disabled.

    Note:
        The StateManager does not autonomously poll hardware state. It relies on the
        service to perform such updates. The cache maintained by
        [`DataServiceCache`][pydase.data_service.data_service_cache.DataServiceCache]
        reflects the last known state as notified by the `DataServiceObserver`, and is
        used by the web interface to provide fast and accurate state rendering for
        connected clients.
    """

    def __init__(
        self,
        service: "DataService",
        filename: str | Path | None = None,
        autosave_interval: float | None = None,
    ) -> None:
        self.filename = getattr(service, "_filename", None)

        if filename is not None:
            if self.filename is not None:
                logger.warning(
                    "Overwriting filename '%s' with '%s'.", self.filename, filename
                )
            self.filename = filename

        self.service = service
        self.cache_manager = DataServiceCache(self.service)
        self.autosave_interval = autosave_interval

    async def autosave(self) -> None:
        """Periodically saves the current service state to the configured file.

        This coroutine is automatically started by the [`pydase.Server`][pydase.Server]
        when a filename is provided. It runs in the background and writes the latest
        known state of the service to disk every `autosave_interval` seconds.

        If `autosave_interval` is set to `None`, autosaving is disabled and this
        coroutine exits immediately.
        """

        if self.autosave_interval is None:
            return

        while True:
            try:
                if self.filename is not None:
                    self.save_state()
                await asyncio.sleep(self.autosave_interval)
            except Exception as e:
                logger.exception(e)

    @property
    def cache_value(self) -> dict[str, SerializedObject]:
        """Returns the "value" value of the DataService serialization."""
        return cast(dict[str, SerializedObject], self.cache_manager.cache["value"])

    def save_state(self) -> None:
        """Saves the DataService's current state to a JSON file defined by
        `self.filename`.
        """

        if self.filename is not None:
            with open(self.filename, "w") as f:
                json.dump(self.cache_value, f, indent=4)
        else:
            logger.debug(
                "State manager was not initialised with a filename. Skipping "
                "'save_state'..."
            )

    def load_state(self) -> None:
        """Loads the DataService's state from a JSON file defined by `self.filename`.
        Updates the service's attributes, respecting type and read-only constraints.
        """

        # Traverse the serialized representation and set the attributes of the class
        json_dict = self._get_state_dict_from_json_file()
        if json_dict == {}:
            logger.debug("Could not load the service state.")
            return

        for path in generate_serialized_data_paths(json_dict):
            if self.__is_loadable_state_attribute(path):
                nested_json_dict = get_nested_dict_by_path(json_dict, path)
                try:
                    nested_class_dict = self.cache_manager.get_value_dict_from_cache(
                        path
                    )
                except (SerializationPathError, KeyError):
                    nested_class_dict = {
                        "full_access_path": path,
                        "value": None,
                        "type": "None",
                        "doc": None,
                        "readonly": False,
                    }

                value_type = nested_json_dict["type"]
                class_attr_value_type = nested_class_dict.get("type", None)

                if class_attr_value_type == value_type:
                    self.set_service_attribute_value_by_path(path, nested_json_dict)
                else:
                    logger.info(
                        "Attribute type of '%s' changed from '%s' to "
                        "'%s'. Ignoring value from JSON file...",
                        path,
                        value_type,
                        class_attr_value_type,
                    )

    def _get_state_dict_from_json_file(self) -> dict[str, Any]:
        if self.filename is not None and os.path.exists(self.filename):
            with open(self.filename) as f:
                # Load JSON data from file and update class attributes with these
                # values
                return cast(dict[str, Any], json.load(f))
        return {}

    def set_service_attribute_value_by_path(
        self,
        path: str,
        serialized_value: SerializedObject,
    ) -> None:
        """Sets the value of an attribute in the service managed by the `StateManager`
        given its path as a dot-separated string.

        This method updates the attribute specified by 'path' with 'value' only if the
        attribute is not read-only and the new value differs from the current one.
        It also handles type-specific conversions for the new value before setting it.

        Args:
            path:
                A dot-separated string indicating the hierarchical path to the
                attribute.
            serialized_value:
                The serialized representation of the new value to set for the attribute.
        """

        try:
            current_value_dict = self.cache_manager.get_value_dict_from_cache(path)
        except (SerializationPathError, KeyError):
            current_value_dict = {
                "full_access_path": path,
                "value": None,
                "type": "None",
                "doc": None,
                "readonly": False,
            }

        if "full_access_path" not in serialized_value:
            # Backwards compatibility for JSON files not containing the
            # full_access_path
            logger.warning(
                "The format of your JSON file is out-of-date. This might lead "
                "to unexpected errors. Please consider updating it."
            )
            serialized_value["full_access_path"] = current_value_dict[
                "full_access_path"
            ]

        # only set value when it has changed
        if self.__attr_value_has_changed(serialized_value, current_value_dict):
            self.__update_attribute_by_path(path, serialized_value)
        else:
            logger.debug("Value of attribute '%s' has not changed...", path)

    def __attr_value_has_changed(
        self, serialized_new_value: Any, serialized_current_value: Any
    ) -> bool:
        return not (
            serialized_new_value["type"] == serialized_current_value["type"]
            and serialized_new_value["value"] == serialized_current_value["value"]
        )

    def __update_attribute_by_path(
        self, path: str, serialized_value: SerializedObject
    ) -> None:
        is_value_set = False
        path_parts = parse_full_access_path(path)
        target_obj = get_object_by_path_parts(self.service, path_parts[:-1])

        if self.__cached_value_is_enum(path):
            enum_attr = get_object_by_path_parts(target_obj, [path_parts[-1]])
            # take the value of the existing enum class
            if serialized_value["type"] in ("ColouredEnum", "Enum"):
                # This error will arise when setting an enum from another enum class.
                # In this case, we resort to loading the enum and setting it directly.
                with contextlib.suppress(KeyError):
                    value = enum_attr.__class__[serialized_value["value"]]
                    is_value_set = True

        if not is_value_set:
            value = loads(serialized_value)

        # set the value
        if isinstance(target_obj, list | dict):
            processed_key = parse_serialized_key(path_parts[-1])
            target_obj[processed_key] = value  # type: ignore
        else:
            # Don't allow adding attributes to objects through state manager
            if self.__attr_exists_on_target_obj(
                target_obj=target_obj, name=path_parts[-1]
            ):
                raise AttributeError(
                    f"{target_obj.__class__.__name__!r} object has no attribute "
                    f"{path_parts[-1]!r}"
                )

            setattr(target_obj, path_parts[-1], value)

    def __is_loadable_state_attribute(self, full_access_path: str) -> bool:
        """Checks if an attribute defined by a dot-separated path should be loaded from
        storage.

        For properties, it verifies the presence of the '@load_state' decorator. Regular
        attributes default to being loadable.
        """

        path_parts = parse_full_access_path(full_access_path)
        parent_object = get_object_by_path_parts(self.service, path_parts[:-1])

        if is_property_attribute(parent_object, path_parts[-1]):
            prop = getattr(type(parent_object), path_parts[-1])
            has_decorator = has_load_state_decorator(prop)
            if not has_decorator:
                logger.debug(
                    "Property '%s' has no '@load_state' decorator. "
                    "Ignoring value from JSON file...",
                    path_parts[-1],
                )
            return has_decorator

        try:
            cached_serialization_dict = self.cache_manager.get_value_dict_from_cache(
                full_access_path
            )

            if cached_serialization_dict["value"] == "method":
                return False

            # nested objects cannot be loaded
            return not serialized_dict_is_nested_object(cached_serialization_dict)
        except SerializationPathError:
            logger.debug(
                "Path %a could not be loaded. It does not correspond to an attribute of"
                " the class. Ignoring value from JSON file...",
                path_parts[-1],
            )
            return False

    def __cached_value_is_enum(self, path: str) -> bool:
        try:
            attr_cache_type = self.cache_manager.get_value_dict_from_cache(path)["type"]

            return attr_cache_type in ("ColouredEnum", "Enum")
        except Exception:
            return False

    def __attr_exists_on_target_obj(self, target_obj: Any, name: str) -> bool:
        return not is_property_attribute(target_obj, name) and not hasattr(
            target_obj, name
        )



import asyncio
import logging
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, cast

import socketio  # type: ignore
from typing_extensions import SupportsIndex

from pydase.utils.serialization.deserializer import Deserializer, loads
from pydase.utils.serialization.serializer import dump
from pydase.utils.serialization.types import SerializedObject

if TYPE_CHECKING:
    from collections.abc import Callable

logger = logging.getLogger(__name__)


class ProxyAttributeError(Exception): ...


def trigger_method(
    sio_client: socketio.AsyncClient,
    loop: asyncio.AbstractEventLoop,
    access_path: str,
    args: list[Any],
    kwargs: dict[str, Any],
) -> Any:
    async def async_trigger_method() -> Any:
        return await sio_client.call(
            "trigger_method",
            {
                "access_path": access_path,
                "args": dump(args),
                "kwargs": dump(kwargs),
            },
        )

    result: SerializedObject | None = asyncio.run_coroutine_threadsafe(
        async_trigger_method(),
        loop=loop,
    ).result()

    if result is not None:
        return ProxyLoader.loads_proxy(
            serialized_object=result, sio_client=sio_client, loop=loop
        )

    return None


def update_value(
    sio_client: socketio.AsyncClient,
    loop: asyncio.AbstractEventLoop,
    access_path: str,
    value: Any,
) -> Any:
    async def set_result() -> Any:
        return await sio_client.call(
            "update_value",
            {
                "access_path": access_path,
                "value": dump(value),
            },
        )

    result: SerializedObject | None = asyncio.run_coroutine_threadsafe(
        set_result(),
        loop=loop,
    ).result()
    if result is not None:
        ProxyLoader.loads_proxy(
            serialized_object=result, sio_client=sio_client, loop=loop
        )


class ProxyDict(dict[str, Any]):
    def __init__(
        self,
        original_dict: dict[str, Any],
        parent_path: str,
        sio_client: socketio.AsyncClient,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        super().__init__(original_dict)
        self._parent_path = parent_path
        self._loop = loop
        self._sio = sio_client

    def __setitem__(self, key: str, value: Any) -> None:
        observer_key = key
        if isinstance(key, str):
            observer_key = f'"{key}"'

        full_access_path = f"{self._parent_path}[{observer_key}]"

        update_value(self._sio, self._loop, full_access_path, value)

    def pop(self, key: str) -> Any:  # type: ignore
        """Removes the element from the dictionary on the server. It does not return
        any proxy as the corresponding object on the server does not live anymore."""

        full_access_path = f"{self._parent_path}.pop"

        trigger_method(self._sio, self._loop, full_access_path, [key], {})


class ProxyList(list[Any]):
    def __init__(
        self,
        original_list: list[Any],
        parent_path: str,
        sio_client: socketio.AsyncClient,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        super().__init__(original_list)
        self._parent_path = parent_path
        self._loop = loop
        self._sio = sio_client

    def __setitem__(self, key: int, value: Any) -> None:  # type: ignore[override]
        full_access_path = f"{self._parent_path}[{key}]"

        update_value(self._sio, self._loop, full_access_path, value)

    def append(self, __object: Any) -> None:
        full_access_path = f"{self._parent_path}.append"

        trigger_method(self._sio, self._loop, full_access_path, [__object], {})

    def clear(self) -> None:
        full_access_path = f"{self._parent_path}.clear"

        trigger_method(self._sio, self._loop, full_access_path, [], {})

    def extend(self, __iterable: Iterable[Any]) -> None:
        full_access_path = f"{self._parent_path}.extend"

        trigger_method(self._sio, self._loop, full_access_path, [__iterable], {})

    def insert(self, __index: SupportsIndex, __object: Any) -> None:
        full_access_path = f"{self._parent_path}.insert"

        trigger_method(self._sio, self._loop, full_access_path, [__index, __object], {})

    def pop(self, __index: SupportsIndex = -1) -> Any:
        full_access_path = f"{self._parent_path}.pop"

        return trigger_method(self._sio, self._loop, full_access_path, [__index], {})

    def remove(self, __value: Any) -> None:
        full_access_path = f"{self._parent_path}.remove"

        trigger_method(self._sio, self._loop, full_access_path, [__value], {})


class ProxyClassMixin:
    def __init__(self) -> None:
        # declare before DataService init to avoid warning messaged
        self._observers: dict[str, Any] = {}

        self._proxy_getters: dict[str, Callable[..., Any]] = {}
        self._proxy_setters: dict[str, Callable[..., Any]] = {}
        self._proxy_methods: dict[str, Callable[..., Any]] = {}

    def _initialise(
        self,
        sio_client: socketio.AsyncClient,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        self._loop = loop
        self._sio = sio_client

    def __dir__(self) -> list[str]:
        """Used to provide tab completion on CLI / notebook"""
        static_dir = super().__dir__()
        return sorted({*static_dir, *self._proxy_getters, *self._proxy_methods.keys()})

    def __getattribute__(self, name: str) -> Any:
        try:
            if name in super().__getattribute__("_proxy_getters"):
                return super().__getattribute__("_proxy_getters")[name]()
            if name in super().__getattribute__("_proxy_methods"):
                return super().__getattribute__("_proxy_methods")[name]
        except AttributeError:
            pass
        return super().__getattribute__(name)

    def __setattr__(self, name: str, value: Any) -> None:
        try:
            if name in super().__getattribute__("_proxy_setters"):
                return super().__getattribute__("_proxy_setters")[name](value)
            if name in super().__getattribute__("_proxy_getters"):
                raise ProxyAttributeError(
                    f"Proxy attribute {name!r} of {type(self).__name__!r} is readonly!"
                )
        except AttributeError:
            pass
        return super().__setattr__(name, value)

    def _handle_serialized_method(
        self, attr_name: str, serialized_object: SerializedObject
    ) -> None:
        if serialized_object["type"] == "method":
            self._add_method_proxy(attr_name, serialized_object)

    def _add_method_proxy(
        self, attr_name: str, serialized_object: SerializedObject
    ) -> None:
        def method_proxy(*args: Any, **kwargs: Any) -> Any:
            return trigger_method(
                self._sio,
                self._loop,
                serialized_object["full_access_path"],
                list(args),
                kwargs,
            )

        dict.__setitem__(self._proxy_methods, attr_name, method_proxy)

    def _add_attr_proxy(
        self, attr_name: str, serialized_object: SerializedObject
    ) -> None:
        self._add_getattr_proxy(attr_name, serialized_object=serialized_object)
        if not serialized_object["readonly"]:
            self._add_setattr_proxy(attr_name, serialized_object=serialized_object)

    def _add_setattr_proxy(
        self, attr_name: str, serialized_object: SerializedObject
    ) -> None:
        self._add_getattr_proxy(attr_name, serialized_object=serialized_object)
        if not serialized_object["readonly"]:

            def setter_proxy(value: Any) -> None:
                update_value(
                    self._sio, self._loop, serialized_object["full_access_path"], value
                )

            dict.__setitem__(self._proxy_setters, attr_name, setter_proxy)  # type: ignore

    def _add_getattr_proxy(
        self, attr_name: str, serialized_object: SerializedObject
    ) -> None:
        def getter_proxy() -> Any:
            async def get_result() -> Any:
                return await self._sio.call(
                    "get_value", serialized_object["full_access_path"]
                )

            result = asyncio.run_coroutine_threadsafe(
                get_result(),
                loop=self._loop,
            ).result()
            return ProxyLoader.loads_proxy(result, self._sio, self._loop)

        dict.__setitem__(self._proxy_getters, attr_name, getter_proxy)  # type: ignore


class ProxyLoader:
    @staticmethod
    def load_list_proxy(
        serialized_object: SerializedObject,
        sio_client: socketio.AsyncClient,
        loop: asyncio.AbstractEventLoop,
    ) -> Any:
        return ProxyList(
            [
                ProxyLoader.loads_proxy(item, sio_client, loop)
                for item in cast(list[SerializedObject], serialized_object["value"])
            ],
            parent_path=serialized_object["full_access_path"],
            sio_client=sio_client,
            loop=loop,
        )

    @staticmethod
    def load_dict_proxy(
        serialized_object: SerializedObject,
        sio_client: socketio.AsyncClient,
        loop: asyncio.AbstractEventLoop,
    ) -> Any:
        return ProxyDict(
            {
                key: ProxyLoader.loads_proxy(value, sio_client, loop)
                for key, value in cast(
                    dict[str, SerializedObject], serialized_object["value"]
                ).items()
            },
            parent_path=serialized_object["full_access_path"],
            sio_client=sio_client,
            loop=loop,
        )

    @staticmethod
    def update_data_service_proxy(
        proxy_class: ProxyClassMixin,
        serialized_object: SerializedObject,
    ) -> Any:
        proxy_class._proxy_getters.clear()
        proxy_class._proxy_setters.clear()
        proxy_class._proxy_methods.clear()
        for key, value in cast(
            dict[str, SerializedObject], serialized_object["value"]
        ).items():
            type_handler: dict[str | None, None | Callable[..., Any]] = {
                None: None,
                "int": proxy_class._add_attr_proxy,
                "float": proxy_class._add_attr_proxy,
                "bool": proxy_class._add_attr_proxy,
                "str": proxy_class._add_attr_proxy,
                "NoneType": proxy_class._add_attr_proxy,
                "Quantity": proxy_class._add_attr_proxy,
                "Enum": proxy_class._add_attr_proxy,
                "ColouredEnum": proxy_class._add_attr_proxy,
                "method": proxy_class._handle_serialized_method,
                "list": proxy_class._add_getattr_proxy,
                "dict": proxy_class._add_getattr_proxy,
            }

            # First go through handled types (as ColouredEnum is also within the
            # components)
            handler = type_handler.get(value["type"])
            if handler:
                handler(key, value)
            else:
                proxy_class._add_getattr_proxy(key, value)

    @staticmethod
    def load_data_service_proxy(
        serialized_object: SerializedObject,
        sio_client: socketio.AsyncClient,
        loop: asyncio.AbstractEventLoop,
    ) -> Any:
        # Custom types like Components or DataService classes
        component_class = cast(
            type, Deserializer.get_service_base_class(serialized_object["type"])
        )
        class_bases = (
            ProxyClassMixin,
            component_class,
        )
        proxy_base_class: type[ProxyClassMixin] = type(
            serialized_object["name"],  # type: ignore
            class_bases,
            {},
        )
        proxy_class_instance = proxy_base_class()
        proxy_class_instance._initialise(sio_client=sio_client, loop=loop)
        ProxyLoader.update_data_service_proxy(
            proxy_class=proxy_class_instance, serialized_object=serialized_object
        )
        return proxy_class_instance

    @staticmethod
    def load_default(
        serialized_object: SerializedObject,
        sio_client: socketio.AsyncClient,
        loop: asyncio.AbstractEventLoop,
    ) -> Any:
        return loads(serialized_object)

    @staticmethod
    def loads_proxy(
        serialized_object: SerializedObject,
        sio_client: socketio.AsyncClient,
        loop: asyncio.AbstractEventLoop,
    ) -> Any:
        type_handler: dict[str | None, None | Callable[..., Any]] = {
            "int": ProxyLoader.load_default,
            "float": ProxyLoader.load_default,
            "bool": ProxyLoader.load_default,
            "str": ProxyLoader.load_default,
            "NoneType": ProxyLoader.load_default,
            "Quantity": ProxyLoader.load_default,
            "Enum": ProxyLoader.load_default,
            "ColouredEnum": ProxyLoader.load_default,
            "Exception": ProxyLoader.load_default,
            "list": ProxyLoader.load_list_proxy,
            "dict": ProxyLoader.load_dict_proxy,
        }

        # First go through handled types (as ColouredEnum is also within the components)
        handler = type_handler.get(serialized_object["type"])
        if handler:
            return handler(
                serialized_object=serialized_object, sio_client=sio_client, loop=loop
            )

        return ProxyLoader.load_data_service_proxy(
            serialized_object=serialized_object, sio_client=sio_client, loop=loop
        )



import logging
from typing import Any

from pydase.observer_pattern.observable.decorators import (
    _validate_value_was_correctly_set,
    has_validate_set_decorator,
)
from pydase.observer_pattern.observable.observable_object import ObservableObject
from pydase.utils.helpers import is_descriptor, is_property_attribute

logger = logging.getLogger(__name__)


class Observable(ObservableObject):
    def __init__(self) -> None:
        super().__init__()
        class_attrs = {
            k: type(self).__dict__[k]
            for k in set(type(self).__dict__)
            - set(Observable.__dict__)
            - set(self.__dict__)
            - {"__annotations__"}
        }
        for name, value in class_attrs.items():
            if isinstance(value, property) or callable(value) or is_descriptor(value):
                # Properties, methods and descriptors have to be stored as class
                # attributes to work properly. So don't make it an instance attribute.
                continue
            self.__dict__[name] = self._initialise_new_objects(name, value)

    def __setattr__(self, name: str, value: Any) -> None:
        if not hasattr(self, "_observers") and name != "_observers":
            logger.warning(
                "Ensure that super().__init__() is called at the start of the '%s' "
                "constructor! Failing to do so may lead to unexpected behavior.",
                type(self).__name__,
            )
            self._observers = {}

        value = self._handle_observable_setattr(name, value)

        super().__setattr__(name, value)

        if is_property_attribute(self, name) and has_validate_set_decorator(
            getattr(type(self), name)
        ):
            _validate_value_was_correctly_set(obj=self, name=name, value=value)
        else:
            self._notify_changed(name, value)

    def __getattribute__(self, name: str) -> Any:
        if is_property_attribute(self, name):
            self._notify_change_start(name)

        value = super().__getattribute__(name)

        if is_property_attribute(self, name):
            # fixes https://github.com/tiqi-group/pydase/issues/187 and
            # https://github.com/tiqi-group/pydase/issues/192
            if isinstance(value, ObservableObject):
                value.add_observer(self, name)
            self._notify_changed(name, value)

        return value

    def _handle_observable_setattr(self, name: str, value: Any) -> Any:
        if name == "_observers":
            return value

        self._remove_observer_if_observable(name)
        value = self._initialise_new_objects(name, value)
        self._notify_change_start(name)
        return value

    def _remove_observer_if_observable(self, name: str) -> None:
        if not is_property_attribute(self, name):
            current_value = getattr(self, name, None)

            if isinstance(current_value, ObservableObject):
                current_value._remove_observer(self, name)

    def _construct_extended_attr_path(
        self, observer_attr_name: str, instance_attr_name: str
    ) -> str:
        if observer_attr_name != "":
            return (
                f"{observer_attr_name}.{instance_attr_name}"
                if instance_attr_name != ""
                else observer_attr_name
            )
        return instance_attr_name



import time
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, ParamSpec, TypeVar

if TYPE_CHECKING:
    from pydase.observer_pattern.observable.observable import Observable

P = ParamSpec("P")
R = TypeVar("R")


def validate_set(
    *, timeout: float = 0.1, precision: float | None = None
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    Decorator marking a property setter to read back the set value using the property
    getter and check against the desired value.

    Args:
        timeout:
            The maximum time (in seconds) to wait for the value to be within the
            precision boundary.
        precision:
            The acceptable deviation from the desired value. If None, the value must be
            exact.
    """

    def validate_set_decorator(func: Callable[P, R]) -> Callable[P, R]:
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            return func(*args, **kwargs)

        wrapper._validate_kwargs = {  # type: ignore
            "timeout": timeout,
            "precision": precision,
        }

        return wrapper

    return validate_set_decorator


def has_validate_set_decorator(prop: property) -> bool:
    """
    Checks if a property setter has been decorated with the `validate_set` decorator.

    Args:
        prop:
            The property to check.

    Returns:
        True if the property setter has the `validate_set` decorator, False otherwise.
    """

    property_setter = prop.fset
    return hasattr(property_setter, "_validate_kwargs")


def _validate_value_was_correctly_set(
    *,
    obj: "Observable",
    name: str,
    value: Any,
) -> None:
    """
    Validates if the property `name` of `obj` attains the desired `value` within the
    specified `precision` and time `timeout`.

    Args:
        obj:
            The instance of the class containing the property.
        name:
            The name of the property to validate.
        value:
            The desired value to check against.

    Raises:
        ValueError:
            If the property value does not match the desired value within the specified
            precision and timeout.
    """

    prop: property = getattr(type(obj), name)

    timeout = prop.fset._validate_kwargs["timeout"]  # type: ignore
    precision = prop.fset._validate_kwargs["precision"]  # type: ignore
    if precision is None:
        precision = 0.0

    start_time = time.time()
    while time.time() - start_time < timeout:
        current_value = obj.__getattribute__(name)
        # This check is faster than rounding and comparing to 0
        if abs(current_value - value) <= precision:
            return
        time.sleep(0.01)
    raise ValueError(
        f"Failed to set value to {value} within {timeout} seconds. Current value: "
        f"{current_value}."
    )



import asyncio
import logging
import os
import signal
from collections.abc import Callable, Coroutine
from datetime import datetime
from time import time
from typing import (
    Generic,
    TypeVar,
)

import pydase.data_service.data_service
from pydase.task.task_status import TaskStatus
from pydase.utils.helpers import current_event_loop_exists

logger = logging.getLogger(__name__)

R = TypeVar("R")


class Task(pydase.data_service.data_service.DataService, Generic[R]):
    """A class representing a task within the `pydase` framework.

    The `Task` class wraps an asynchronous function and provides methods to manage its
    lifecycle, such as `start()` and `stop()`. It is typically used to perform periodic
    or recurring jobs in a [`DataService`][pydase.DataService], like reading
    sensor data, updating databases, or executing other background tasks.

    When a function is decorated with the [`@task`][pydase.task.decorator.task]
    decorator, it is replaced by a `Task` instance that controls the execution of the
    original function.

    The keyword arguments that can be passed to this class are inspired by systemd unit
    services.

    Args:
        func:
            The asynchronous function that this task wraps. It must be a coroutine
            without arguments.
        autostart:
            If set to True, the task will automatically start when the service is
            initialized. Defaults to False.
        restart_on_exception:
            Configures whether the task shall be restarted when it exits with an
            exception other than [`asyncio.CancelledError`][asyncio.CancelledError].
        restart_sec:
            Configures the time to sleep before restarting a task. Defaults to 1.0.
        start_limit_interval_sec:
            Configures start rate limiting. Tasks which are started more than
            `start_limit_burst` times within an `start_limit_interval_sec` time span are
            not permitted to start any more. Defaults to None (disabled rate limiting).
        start_limit_burst:
            Configures unit start rate limiting. Tasks which are started more than
            `start_limit_burst` times within an `start_limit_interval_sec` time span are
            not permitted to start any more. Defaults to 3.
        exit_on_failure:
            If True, exit the service if the task fails and restart_on_exception is
            False or burst limits are exceeded.

    Example:
        ```python
        import asyncio

        import pydase
        from pydase.task.decorator import task


        class MyService(pydase.DataService):
            @task(autostart=True)
            async def my_task(self) -> None:
                while True:
                    # Perform some periodic work
                    await asyncio.sleep(1)


        if __name__ == "__main__":
            service = MyService()
            pydase.Server(service=service).run()
        ```

        In this example, `my_task` is defined as a task using the `@task` decorator, and
        it will start automatically when the service is initialized because
        `autostart=True` is set. You can manually start or stop the task using
        `service.my_task.start()` and `service.my_task.stop()`, respectively.
    """

    def __init__(  # noqa: PLR0913
        self,
        func: Callable[[], Coroutine[None, None, R | None]],
        *,
        autostart: bool,
        restart_on_exception: bool,
        restart_sec: float,
        start_limit_interval_sec: float | None,
        start_limit_burst: int,
        exit_on_failure: bool,
    ) -> None:
        super().__init__()
        self._autostart = autostart
        self._restart_on_exception = restart_on_exception
        self._restart_sec = restart_sec
        self._start_limit_interval_sec = start_limit_interval_sec
        self._start_limit_burst = start_limit_burst
        self._exit_on_failure = exit_on_failure
        self._func_name = func.__name__
        self._func = func
        self._task: asyncio.Task[R | None] | None = None
        self._status = TaskStatus.NOT_RUNNING
        self._result: R | None = None

        if not current_event_loop_exists():
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)
        else:
            self._loop = asyncio.get_event_loop()

    @property
    def autostart(self) -> bool:
        """Defines if the task should be started automatically when the
        [`Server`][pydase.Server] starts."""
        return self._autostart

    @property
    def status(self) -> TaskStatus:
        """Returns the current status of the task."""
        return self._status

    def start(self) -> None:
        """Starts the asynchronous task if it is not already running."""
        if self._task:
            return

        def task_done_callback(task: asyncio.Task[R | None]) -> None:
            """Handles tasks that have finished.

            Updates the task status, calls the defined callbacks, and logs and re-raises
            exceptions.
            """

            self._task = None
            self._status = TaskStatus.NOT_RUNNING

            exception = None
            try:
                exception = task.exception()
            except asyncio.CancelledError:
                return

            if exception is not None:
                logger.error(
                    "Task '%s' encountered an exception: %r",
                    self._func_name,
                    exception,
                )
                os.kill(os.getpid(), signal.SIGTERM)
            else:
                self._result = task.result()

        logger.info("Creating task %r", self._func_name)
        self._task = self._loop.create_task(self.__running_task_loop())
        self._task.add_done_callback(task_done_callback)

    async def __running_task_loop(self) -> R | None:
        logger.info("Starting task %r", self._func_name)
        self._status = TaskStatus.RUNNING
        attempts = 0
        start_time_of_start_limit_interval = None

        while True:
            try:
                return await self._func()
            except asyncio.CancelledError:
                logger.info("Task '%s' was cancelled", self._func_name)
                raise
            except Exception as e:
                attempts, start_time_of_start_limit_interval = (
                    self._handle_task_exception(
                        e, attempts, start_time_of_start_limit_interval
                    )
                )
                if not self._should_restart_task(
                    attempts, start_time_of_start_limit_interval
                ):
                    if self._exit_on_failure:
                        raise e
                    break
                await asyncio.sleep(self._restart_sec)
        return None

    def _handle_task_exception(
        self,
        exception: Exception,
        attempts: int,
        start_time_of_start_limit_interval: float | None,
    ) -> tuple[int, float]:
        """Handle an exception raised during task execution."""
        if start_time_of_start_limit_interval is None:
            start_time_of_start_limit_interval = time()

        attempts += 1
        logger.exception(
            "Task %r encountered an exception: %r [attempt %s since %s].",
            self._func.__name__,
            exception,
            attempts,
            datetime.fromtimestamp(start_time_of_start_limit_interval),
        )
        return attempts, start_time_of_start_limit_interval

    def _should_restart_task(
        self, attempts: int, start_time_of_start_limit_interval: float
    ) -> bool:
        """Determine if the task should be restarted."""
        if not self._restart_on_exception:
            return False

        if self._start_limit_interval_sec is not None:
            if (
                time() - start_time_of_start_limit_interval
            ) > self._start_limit_interval_sec:
                # Reset attempts if interval is exceeded
                start_time_of_start_limit_interval = time()
                attempts = 1
            elif attempts > self._start_limit_burst:
                logger.error(
                    "Task %r exceeded restart burst limit. Stopping.",
                    self._func.__name__,
                )
                return False
        return True

    def stop(self) -> None:
        """Stops the running asynchronous task by cancelling it."""

        if self._task:
            self._task.cancel()



import asyncio
import html
import json
import logging
from pathlib import Path
from typing import Any

import aiohttp.web
import aiohttp_middlewares.cors
import anyio

from pydase.config import ServiceConfig, WebServerConfig
from pydase.data_service.data_service_observer import DataServiceObserver
from pydase.server.web_server.api import create_api_application
from pydase.server.web_server.sio_setup import (
    setup_sio_server,
)
from pydase.utils.helpers import (
    get_path_from_path_parts,
    parse_full_access_path,
)
from pydase.utils.serialization.serializer import generate_serialized_data_paths

logger = logging.getLogger(__name__)


class WebServer:
    """
    Represents a web server that adheres to the
    [`AdditionalServerProtocol`][pydase.server.server.AdditionalServerProtocol],
    designed to work with a [`DataService`][pydase.DataService] instance. This server
    facilitates client-server communication and state management through web protocols
    and socket connections.

    The WebServer class initializes and manages a web server environment aiohttp and
    Socket.IO, allowing for HTTP and Socket.IO communications. It incorporates CORS
    (Cross-Origin Resource Sharing) support, custom CSS, and serves a static files
    directory. It also initializes web server settings based on configuration files or
    generates default settings if necessary.

    Configuration for the web server (like service configuration directory and whether
    to generate new web settings) is determined in the following order of precedence:

    1. Values provided directly to the constructor.
    2. Environment variable settings (via configuration classes like
      [`ServiceConfig`][pydase.config.ServiceConfig] and
      [`WebServerConfig`][pydase.config.WebServerConfig]).
    3. Default values defined in the configuration classes.

    Args:
        data_service_observer:
            Observer for the [`DataService`][pydase.DataService], handling state updates
            and communication to connected clients.
        host:
            Hostname or IP address where the server is accessible. Commonly '0.0.0.0'
            to bind to all network interfaces.
        port:
            Port number on which the server listens. Typically in the range 1024-65535
            (non-standard ports).
        css:
            Path to a custom CSS file for styling the frontend. If None, no custom
            styles are applied. Defaults to None.
        favicon_path:
            Path to a custom favicon.ico file. Defaults to None.
        enable_cors:
            Flag to enable or disable CORS policy. When True, CORS is enabled, allowing
            cross-origin requests. Defaults to True.
        config_dir:
            Path to the configuration directory where the web settings will be stored.
            Defaults to
            [`ServiceConfig().config_dir`][pydase.config.ServiceConfig.config_dir].
        generate_web_settings:
            Flag to enable or disable generation of new web settings if the
            configuration file is missing. Defaults to
            [`WebServerConfig().generate_web_settings`][pydase.config.WebServerConfig.generate_web_settings].
    """

    def __init__(  # noqa: PLR0913
        self,
        data_service_observer: DataServiceObserver,
        host: str,
        port: int,
        *,
        css: str | Path | None = None,
        favicon_path: str | Path | None = None,
        enable_cors: bool = True,
        config_dir: Path = ServiceConfig().config_dir,
        generate_web_settings: bool = WebServerConfig().generate_web_settings,
        frontend_src: Path = Path(__file__).parent.parent.parent / "frontend",
    ) -> None:
        self.observer = data_service_observer
        self.state_manager = self.observer.state_manager
        self.service = self.state_manager.service
        self.port = port
        self.host = host
        self.css = css
        self.enable_cors = enable_cors
        self.frontend_src = frontend_src
        self.favicon_path: Path | str = favicon_path  # type: ignore

        if self.favicon_path is None:
            self.favicon_path = self.frontend_src / "favicon.ico"

        self._service_config_dir = config_dir
        self._generate_web_settings = generate_web_settings
        self._loop: asyncio.AbstractEventLoop
        self._initialise_configuration()

    async def serve(self) -> None:
        self._loop = asyncio.get_running_loop()
        self._sio = setup_sio_server(self.observer, self.enable_cors, self._loop)

        async def index(
            request: aiohttp.web.Request,
        ) -> aiohttp.web.Response | aiohttp.web.FileResponse:
            forwarded_proto = request.headers.get("X-Forwarded-Proto", "http")
            escaped_proto = html.escape(forwarded_proto)

            # Read the index.html file
            index_file_path = self.frontend_src / "index.html"

            async with await anyio.open_file(index_file_path) as f:
                html_content = await f.read()

            # Inject the escaped forwarded protocol into the HTML
            modified_html = html_content.replace(
                'window.__FORWARDED_PROTO__ = "";',
                f'window.__FORWARDED_PROTO__ = "{escaped_proto}";',
            )

            # Read the X-Forwarded-Prefix header from the request
            forwarded_prefix = request.headers.get("X-Forwarded-Prefix", "")

            if forwarded_prefix != "":
                # Escape the forwarded prefix to prevent XSS
                escaped_prefix = html.escape(forwarded_prefix)

                # Inject the escaped forwarded prefix into the HTML
                modified_html = modified_html.replace(
                    'window.__FORWARDED_PREFIX__ = "";',
                    f'window.__FORWARDED_PREFIX__ = "{escaped_prefix}";',
                )
                modified_html = modified_html.replace(
                    "/assets/",
                    f"{escaped_prefix}/assets/",
                )

                modified_html = modified_html.replace(
                    "/favicon.ico",
                    f"{escaped_prefix}/favicon.ico",
                )

            return aiohttp.web.Response(text=modified_html, content_type="text/html")

        app = aiohttp.web.Application()

        # Add CORS middleware if enabled
        if self.enable_cors:
            app.middlewares.append(
                aiohttp_middlewares.cors.cors_middleware(allow_all=True)
            )

        # Define routes
        self._sio.attach(app, socketio_path="/ws/socket.io")
        app.router.add_static("/assets", self.frontend_src / "assets")
        app.router.add_get("/favicon.ico", self._favicon_route)
        app.router.add_get("/service-properties", self._service_properties_route)
        app.router.add_get("/web-settings", self._web_settings_route)
        app.router.add_get("/custom.css", self._styles_route)
        app.add_subapp("/api/", create_api_application(self.state_manager))

        app.router.add_get(r"/", index)
        app.router.add_get(r"/{tail:.*}", index)

        await aiohttp.web._run_app(
            app,
            host=self.host,
            port=self.port,
            handle_signals=False,
            print=logger.info,
            shutdown_timeout=0.1,
        )

    async def _favicon_route(
        self,
        request: aiohttp.web.Request,
    ) -> aiohttp.web.FileResponse:
        return aiohttp.web.FileResponse(self.favicon_path)

    async def _service_properties_route(
        self,
        request: aiohttp.web.Request,
    ) -> aiohttp.web.Response:
        return aiohttp.web.json_response(self.state_manager.cache_manager.cache)

    async def _web_settings_route(
        self,
        request: aiohttp.web.Request,
    ) -> aiohttp.web.Response:
        return aiohttp.web.json_response(self.web_settings)

    async def _styles_route(
        self,
        request: aiohttp.web.Request,
    ) -> aiohttp.web.FileResponse | aiohttp.web.Response:
        if self.css is not None:
            return aiohttp.web.FileResponse(self.css)

        return aiohttp.web.Response(content_type="text/css")

    def _initialise_configuration(self) -> None:
        logger.debug("Initialising web server configuration...")

        if self._generate_web_settings:
            logger.debug("Generating web settings file...")
            file_path = self._service_config_dir / "web_settings.json"

            # File does not exist, create it with default content
            file_path.parent.mkdir(
                parents=True, exist_ok=True
            )  # Ensure directory exists
            file_path.write_text(json.dumps(self.web_settings, indent=4))

    def _get_web_settings_from_file(self) -> dict[str, dict[str, Any]]:
        file_path = self._service_config_dir / "web_settings.json"
        web_settings = {}

        # File exists, read its content
        if file_path.exists():
            logger.debug(
                "Reading configuration from file '%s' ...", file_path.absolute()
            )

            web_settings = json.loads(file_path.read_text())

        return web_settings

    @property
    def web_settings(self) -> dict[str, dict[str, Any]]:
        current_web_settings = self._get_web_settings_from_file()
        for path in generate_serialized_data_paths(self.state_manager.cache_value):
            if path in current_web_settings:
                continue

            # Creating the display name by reversely looping through the path parts
            # until an item does not start with a square bracket, and putting the parts
            # back together again. This allows for display names like
            #       >>> 'dict_attr["some.dotted.key"]'
            display_name_parts: list[str] = []
            for item in parse_full_access_path(path)[::-1]:
                display_name_parts.insert(0, item)
                if not item.startswith("["):
                    break

            current_web_settings[path] = {
                "displayName": get_path_from_path_parts(display_name_parts),
                "display": True,
            }

        return current_web_settings



import logging
from collections.abc import Callable, Coroutine
from typing import Any, Generic, TypeVar, overload

from pydase.data_service.data_service import DataService
from pydase.task.task import Task

logger = logging.getLogger(__name__)

R = TypeVar("R")


class PerInstanceTaskDescriptor(Generic[R]):
    """
    A descriptor class that provides a unique [`Task`][pydase.task.task.Task] object
    for each instance of a [`DataService`][pydase.data_service.data_service.DataService]
    class.

    The `PerInstanceTaskDescriptor` is used to transform an asynchronous function into a
    task that is managed independently for each instance of a `DataService` subclass.
    This allows tasks to be initialized, started, and stopped on a per-instance basis,
    providing better control over task execution within the service.

    The `PerInstanceTaskDescriptor` is not intended to be used directly. Instead, it is
    used internally by the `@task` decorator to manage task objects for each instance of
    the service class.
    """

    def __init__(  # noqa: PLR0913
        self,
        func: Callable[[Any], Coroutine[None, None, R]]
        | Callable[[], Coroutine[None, None, R]],
        autostart: bool,
        restart_on_exception: bool,
        restart_sec: float,
        start_limit_interval_sec: float | None,
        start_limit_burst: int,
        exit_on_failure: bool,
    ) -> None:
        self.__func = func
        self.__autostart = autostart
        self.__task_instances: dict[object, Task[R]] = {}
        self.__restart_on_exception = restart_on_exception
        self.__restart_sec = restart_sec
        self.__start_limit_interval_sec = start_limit_interval_sec
        self.__start_limit_burst = start_limit_burst
        self.__exit_on_failure = exit_on_failure

    def __set_name__(self, owner: type[DataService], name: str) -> None:
        """Stores the name of the task within the owning class. This method is called
        automatically when the descriptor is assigned to a class attribute.
        """

        self.__task_name = name

    @overload
    def __get__(
        self, instance: None, owner: type[DataService]
    ) -> "PerInstanceTaskDescriptor[R]":
        """Returns the descriptor itself when accessed through the class."""

    @overload
    def __get__(self, instance: DataService, owner: type[DataService]) -> Task[R]:
        """Returns the `Task` object associated with the specific `DataService`
        instance.
        If no task exists for the instance, a new `Task` object is created and stored
        in the `__task_instances` dictionary.
        """

    def __get__(
        self, instance: DataService | None, owner: type[DataService]
    ) -> "Task[R] | PerInstanceTaskDescriptor[R]":
        if instance is None:
            return self

        # Create a new Task object for this instance, using the function's name.
        if instance not in self.__task_instances:
            self.__task_instances[instance] = instance._initialise_new_objects(
                self.__task_name,
                Task(
                    self.__func.__get__(instance, owner),
                    autostart=self.__autostart,
                    restart_on_exception=self.__restart_on_exception,
                    restart_sec=self.__restart_sec,
                    start_limit_interval_sec=self.__start_limit_interval_sec,
                    start_limit_burst=self.__start_limit_burst,
                    exit_on_failure=self.__exit_on_failure,
                ),
            )

        return self.__task_instances[instance]


def task(  # noqa: PLR0913
    *,
    autostart: bool = False,
    restart_on_exception: bool = True,
    restart_sec: float = 1.0,
    start_limit_interval_sec: float | None = None,
    start_limit_burst: int = 3,
    exit_on_failure: bool = False,
) -> Callable[
    [
        Callable[[Any], Coroutine[None, None, R]]
        | Callable[[], Coroutine[None, None, R]]
    ],
    PerInstanceTaskDescriptor[R],
]:
    """
    A decorator to define an asynchronous function as a per-instance task within a
    [`DataService`][pydase.DataService] class.

    This decorator transforms an asynchronous function into a
    [`Task`][pydase.task.task.Task] object that is unique to each instance of the
    `DataService` class. The resulting `Task` object provides methods like `start()`
    and `stop()` to control the execution of the task, and manages the task's lifecycle
    independently for each instance of the service.

    The decorator is particularly useful for defining tasks that need to run
    periodically or perform asynchronous operations, such as polling data sources,
    updating databases, or any recurring job that should be managed within the context
    of a `DataService`.

    The keyword arguments that can be passed to this decorator are inspired by systemd
    unit services.

    Args:
        autostart:
            If set to True, the task will automatically start when the service is
            initialized. Defaults to False.
        restart_on_exception:
            Configures whether the task shall be restarted when it exits with an
            exception other than [`asyncio.CancelledError`][asyncio.CancelledError].
        restart_sec:
            Configures the time to sleep before restarting a task. Defaults to 1.0.
        start_limit_interval_sec:
            Configures start rate limiting. Tasks which are started more than
            `start_limit_burst` times within an `start_limit_interval_sec` time span are
            not permitted to start any more. Defaults to None (disabled rate limiting).
        start_limit_burst:
            Configures unit start rate limiting. Tasks which are started more than
            `start_limit_burst` times within an `start_limit_interval_sec` time span are
            not permitted to start any more. Defaults to 3.
        exit_on_failure:
            If True, exit the service if the task fails and restart_on_exception is
            False or burst limits are exceeded.
    Returns:
        A decorator that wraps an asynchronous function in a
        [`PerInstanceTaskDescriptor`][pydase.task.decorator.PerInstanceTaskDescriptor]
        object, which, when accessed, provides an instance-specific
        [`Task`][pydase.task.task.Task] object.

    Example:
        ```python
        import asyncio

        import pydase
        from pydase.task.decorator import task


        class MyService(pydase.DataService):
            @task(autostart=True)
            async def my_task(self) -> None:
                while True:
                    # Perform some periodic work
                    await asyncio.sleep(1)


        if __name__ == "__main__":
            service = MyService()
            pydase.Server(service=service).run()
        ```

        In this example, `my_task` is defined as a task using the `@task` decorator, and
        it will start automatically when the service is initialized because
        `autostart=True` is set. You can manually start or stop the task using
        `service.my_task.start()` and `service.my_task.stop()`, respectively.
    """

    def decorator(
        func: Callable[[Any], Coroutine[None, None, R]]
        | Callable[[], Coroutine[None, None, R]],
    ) -> PerInstanceTaskDescriptor[R]:
        return PerInstanceTaskDescriptor(
            func,
            autostart=autostart,
            restart_on_exception=restart_on_exception,
            restart_sec=restart_sec,
            start_limit_interval_sec=start_limit_interval_sec,
            start_limit_burst=start_limit_burst,
            exit_on_failure=exit_on_failure,
        )

    return decorator



from pydase.client.client import Client
from pydase.data_service import DataService
from pydase.server import Server
from pydase.utils.logging import setup_logging

setup_logging()

__all__ = [
    "DataService",
    "Server",
    "Client",
]



from __future__ import annotations

import inspect
import logging
import sys
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Any, Literal, cast

import pydase.units as u
from pydase.data_service.abstract_data_service import AbstractDataService
from pydase.task.task_status import TaskStatus
from pydase.utils.decorators import render_in_frontend
from pydase.utils.helpers import (
    get_attribute_doc,
    get_component_classes,
    get_data_service_class_reference,
    get_task_class,
    is_property_attribute,
    parse_full_access_path,
    parse_serialized_key,
)
from pydase.utils.serialization.types import (
    DataServiceTypes,
    SerializedBool,
    SerializedDataService,
    SerializedDatetime,
    SerializedDict,
    SerializedEnum,
    SerializedException,
    SerializedFloat,
    SerializedInteger,
    SerializedList,
    SerializedMethod,
    SerializedNoneType,
    SerializedObject,
    SerializedQuantity,
    SerializedString,
    SignatureDict,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from pydase.client.proxy_class import ProxyClass

logger = logging.getLogger(__name__)


class SerializationError(Exception):
    pass


class SerializationPathError(Exception):
    pass


class Serializer:
    """Serializes objects into
    [`SerializedObject`][pydase.utils.serialization.types.SerializedObject]
    representations.
    """

    @classmethod
    def serialize_object(cls, obj: Any, access_path: str = "") -> SerializedObject:  # noqa: C901
        """Serialize `obj` to a
        [`SerializedObject`][pydase.utils.serialization.types.SerializedObject].

        Args:
            obj:
                Object to be serialized.
            access_path:
                String corresponding to the full access path of the object. This will be
                prepended to the full_access_path in the SerializedObject entries.

        Returns:
            Dictionary representation of `obj`.
        """
        from pydase.client.client import ProxyClass

        result: SerializedObject

        if isinstance(obj, Exception):
            result = cls._serialize_exception(obj)

        elif isinstance(obj, datetime):
            result = cls._serialize_datetime(obj, access_path=access_path)

        elif isinstance(obj, ProxyClass):
            result = cls._serialize_proxy_class(obj, access_path=access_path)

        elif isinstance(obj, AbstractDataService):
            result = cls._serialize_data_service(obj, access_path=access_path)

        elif isinstance(obj, list):
            result = cls._serialize_list(obj, access_path=access_path)

        elif isinstance(obj, dict):
            result = cls._serialize_dict(obj, access_path=access_path)

        # Special handling for u.Quantity
        elif isinstance(obj, u.Quantity):
            result = cls._serialize_quantity(obj, access_path=access_path)

        # Handling for Enums
        elif isinstance(obj, Enum):
            result = cls._serialize_enum(obj, access_path=access_path)

        # Methods and coroutines
        elif inspect.isfunction(obj) or inspect.ismethod(obj):
            result = cls._serialize_method(obj, access_path=access_path)

        elif isinstance(obj, int | float | bool | str | None):
            result = cls._serialize_primitive(obj, access_path=access_path)

        try:
            return result
        except UnboundLocalError:
            raise SerializationError(
                f"Could not serialized object of type {type(obj)}."
            )

    @classmethod
    def _serialize_primitive(
        cls,
        obj: float | bool | str | None,
        access_path: str,
    ) -> (
        SerializedInteger
        | SerializedFloat
        | SerializedBool
        | SerializedString
        | SerializedNoneType
    ):
        doc = get_attribute_doc(obj)
        return {  # type: ignore
            "full_access_path": access_path,
            "doc": doc,
            "readonly": False,
            "type": type(obj).__name__,
            "value": obj,
        }

    @classmethod
    def _serialize_datetime(cls, obj: datetime, access_path: str) -> SerializedDatetime:
        return {
            "type": "datetime",
            "value": str(obj),
            "doc": None,
            "full_access_path": access_path,
            "readonly": True,
        }

    @classmethod
    def _serialize_exception(cls, obj: Exception) -> SerializedException:
        return {
            "full_access_path": "",
            "doc": None,
            "readonly": True,
            "type": "Exception",
            "value": obj.args[0],
            "name": obj.__class__.__name__,
        }

    @classmethod
    def _serialize_enum(cls, obj: Enum, access_path: str = "") -> SerializedEnum:
        import pydase.components.coloured_enum

        value = obj.name
        doc = obj.__doc__
        class_name = type(obj).__name__
        if sys.version_info < (3, 11) and doc == "An enumeration.":
            doc = None
        if isinstance(obj, pydase.components.coloured_enum.ColouredEnum):
            obj_type: Literal["ColouredEnum", "Enum"] = "ColouredEnum"
        else:
            obj_type = "Enum"

        return {
            "full_access_path": access_path,
            "name": class_name,
            "type": obj_type,
            "value": value,
            "readonly": False,
            "doc": doc,
            "enum": {
                name: member.value for name, member in obj.__class__.__members__.items()
            },
        }

    @classmethod
    def _serialize_quantity(
        cls, obj: u.Quantity, access_path: str = ""
    ) -> SerializedQuantity:
        doc = get_attribute_doc(obj)
        value: u.QuantityDict = {"magnitude": obj.m, "unit": str(obj.u)}
        return {
            "full_access_path": access_path,
            "type": "Quantity",
            "value": value,
            "readonly": False,
            "doc": doc,
        }

    @classmethod
    def _serialize_dict(
        cls, obj: dict[str, Any], access_path: str = ""
    ) -> SerializedDict:
        readonly = False
        doc = get_attribute_doc(obj)
        value = {}
        for key, val in obj.items():
            value[key] = cls.serialize_object(
                val, access_path=f'{access_path}["{key}"]'
            )
        return {
            "full_access_path": access_path,
            "type": "dict",
            "value": value,
            "readonly": readonly,
            "doc": doc,
        }

    @classmethod
    def _serialize_list(cls, obj: list[Any], access_path: str = "") -> SerializedList:
        readonly = False
        doc = get_attribute_doc(obj)
        value = [
            cls.serialize_object(o, access_path=f"{access_path}[{i}]")
            for i, o in enumerate(obj)
        ]
        return {
            "full_access_path": access_path,
            "type": "list",
            "value": value,
            "readonly": readonly,
            "doc": doc,
        }

    @classmethod
    def _serialize_method(
        cls, obj: Callable[..., Any], access_path: str = ""
    ) -> SerializedMethod:
        readonly = True
        doc = get_attribute_doc(obj)
        frontend_render = render_in_frontend(obj)

        # Store parameters and their anotations in a dictionary
        sig = inspect.signature(obj)
        sig.return_annotation

        signature: SignatureDict = {"parameters": {}, "return_annotation": {}}

        for k, v in sig.parameters.items():
            default_value = cast(
                dict[str, Any], {} if v.default == inspect._empty else dump(v.default)
            )
            default_value.pop("full_access_path", None)
            signature["parameters"][k] = {
                "annotation": str(v.annotation),
                "default": default_value,
            }

        return {
            "full_access_path": access_path,
            "type": "method",
            "value": None,
            "readonly": readonly,
            "doc": doc,
            "async": inspect.iscoroutinefunction(obj),
            "signature": signature,
            "frontend_render": frontend_render,
        }

    @classmethod
    def _serialize_data_service(
        cls, obj: AbstractDataService, access_path: str = ""
    ) -> SerializedDataService:
        readonly = False
        doc = get_attribute_doc(obj)
        obj_type: DataServiceTypes = "DataService"
        obj_name = obj.__class__.__name__

        # Get component base class if any
        component_base_cls = next(
            (cls for cls in get_component_classes() if isinstance(obj, cls)), None
        )
        if component_base_cls:
            obj_type = component_base_cls.__name__  # type: ignore

        elif isinstance(obj, get_task_class()):
            # Check if obj is a pydase task
            obj_type = "Task"

        # Get the set of DataService class attributes
        data_service_attr_set = set(dir(get_data_service_class_reference()))
        # Get the set of the object attributes
        obj_attr_set = set(dir(obj))
        # Get the difference between the two sets
        derived_only_attr_set = obj_attr_set - data_service_attr_set

        value: dict[str, SerializedObject] = {}

        # Iterate over attributes, properties, class attributes, and methods
        for key in sorted(derived_only_attr_set):
            if key.startswith("_"):
                continue  # Skip attributes that start with underscore

            val = getattr(obj, key)

            path = f"{access_path}.{key}" if access_path else key
            serialized_object = cls.serialize_object(val, access_path=path)

            value[key] = serialized_object

            # If the DataService attribute is a property
            if is_property_attribute(obj, key):
                prop: property = getattr(obj.__class__, key)
                value[key]["readonly"] = prop.fset is None
                value[key]["doc"] = get_attribute_doc(prop)  # overwrite the doc

        return {
            "full_access_path": access_path,
            "name": obj_name,
            "type": obj_type,
            "value": value,
            "readonly": readonly,
            "doc": doc,
        }

    @classmethod
    def _serialize_proxy_class(
        cls, obj: ProxyClass, access_path: str = ""
    ) -> SerializedDataService:
        # Get serialization value from the remote service and adapt the full_access_path
        return add_prefix_to_full_access_path(obj.serialize(), access_path)


def dump(obj: Any) -> SerializedObject:
    """Serialize `obj` to a
    [`SerializedObject`][pydase.utils.serialization.types.SerializedObject].

    The [`Serializer`][pydase.utils.serialization.serializer.Serializer] is used for
    encoding.

    Args:
        obj:
            Object to be serialized.

    Returns:
        Dictionary representation of `obj`.
    """
    return Serializer.serialize_object(obj)


def set_nested_value_by_path(
    serialization_dict: dict[Any, SerializedObject], path: str, value: Any
) -> None:
    """
    Set a value in a nested dictionary structure, which conforms to the serialization
    format used by [`Serializer`][pydase.utils.serialization.serializer.Serializer],
    using a dot-notation path.

    Args:
        serialization_dict:
            The base dictionary representing data serialized with
            [`Serializer`][pydase.utils.serialization.serializer.Serializer].
        path:
            The dot-notation path (e.g., 'attr1.attr2[0].attr3') indicating where to
            set the value.
        value:
            The new value to set at the specified path.

    Note:
        If the index equals the length of the list, the function will append the
        serialized representation of the 'value' to the list.
    """

    path_parts = parse_full_access_path(path)
    current_dict: dict[Any, SerializedObject] = serialization_dict

    try:
        for path_part in path_parts[:-1]:
            next_level_serialized_object = get_container_item_by_key(
                current_dict, path_part, allow_append=False
            )
            current_dict = cast(
                dict[Any, SerializedObject],
                next_level_serialized_object["value"],
            )

        next_level_serialized_object = get_container_item_by_key(
            current_dict, path_parts[-1], allow_append=True
        )
    except (SerializationPathError, KeyError) as e:
        logger.exception("Error occured trying to change %a: %s", path, e)
        return

    if next_level_serialized_object["type"] == "method":  # state change of task
        next_level_serialized_object["value"] = (
            "RUNNING" if isinstance(value, TaskStatus) else None
        )
    else:
        serialized_value = Serializer.serialize_object(value, access_path=path)
        serialized_value["readonly"] = next_level_serialized_object["readonly"]

        keys_to_keep = set(serialized_value.keys())

        next_level_serialized_object.update(serialized_value)  # type: ignore

        # removes keys that are not present in the serialized new value
        for key in list(next_level_serialized_object.keys()):
            if key not in keys_to_keep:
                next_level_serialized_object.pop(key, None)  # type: ignore


def get_nested_dict_by_path(
    serialization_dict: dict[Any, SerializedObject],
    path: str,
) -> SerializedObject:
    path_parts = parse_full_access_path(path)
    current_dict: dict[Any, SerializedObject] = serialization_dict

    for path_part in path_parts[:-1]:
        next_level_serialized_object = get_container_item_by_key(
            current_dict, path_part, allow_append=False
        )
        current_dict = cast(
            dict[Any, SerializedObject],
            next_level_serialized_object["value"],
        )
    return get_container_item_by_key(current_dict, path_parts[-1], allow_append=False)


def create_empty_serialized_object() -> SerializedObject:
    """Create a new empty serialized object."""

    return {
        "full_access_path": "",
        "value": None,
        "type": "None",
        "doc": None,
        "readonly": False,
    }


def get_or_create_item_in_container(
    container: dict[Any, SerializedObject] | list[SerializedObject],
    key: Any,
    *,
    allow_add_key: bool,
) -> SerializedObject:
    """Ensure the key exists in the dictionary, append if necessary and allowed."""

    try:
        return container[key]
    except IndexError:
        if allow_add_key and key == len(container):
            cast(list[SerializedObject], container).append(
                create_empty_serialized_object()
            )
            return container[key]
        raise
    except KeyError:
        if allow_add_key:
            container[key] = create_empty_serialized_object()
            return container[key]
        raise


def get_container_item_by_key(
    container: dict[Any, SerializedObject] | list[SerializedObject],
    key: str,
    *,
    allow_append: bool = False,
) -> SerializedObject:
    """
    Retrieve an item from a container specified by the passed key. Add an item to the
    container if `allow_append` is set to `True`.

    If specified keys or indexes do not exist, the function can append new elements to
    dictionaries and to lists if `allow_append` is True and the missing element is
    exactly the next sequential index (for lists).

    Args:
        container:
            The container representing serialized data.
        key:
            The key name representing the attribute in the dictionary, which may include
            direct keys or indexes (e.g., 'attr_name', '["key"]' or '[0]').
        allow_append:
            Flag to allow appending a new entry if the specified index is out of range
            by exactly one position.

    Returns:
        The dictionary or list item corresponding to the specified attribute and index.

    Raises:
        SerializationPathError:
            If the path composed of `attr_name` and any specified index is invalid, or
            leads to an IndexError or KeyError. This error is also raised if an attempt
            to access a nonexistent key or index occurs without permission to append.
    """
    processed_key = parse_serialized_key(key)

    try:
        return get_or_create_item_in_container(
            container, processed_key, allow_add_key=allow_append
        )
    except IndexError as e:
        raise SerializationPathError(f"Index '{processed_key}': {e}")
    except KeyError as e:
        raise SerializationPathError(f"Key '{processed_key}': {e}")


def get_data_paths_from_serialized_object(  # noqa: C901
    serialized_obj: SerializedObject,
    parent_path: str = "",
) -> list[str]:
    """
    Recursively extracts full access paths from a serialized object.

    Args:
        serialized_obj:
            The dictionary representing the serialization of an object. Produced by
            `pydase.utils.serializer.Serializer`.

    Returns:
        A list of strings, each representing a full access path in the serialized
            object.
    """

    paths: list[str] = []

    if isinstance(serialized_obj["value"], list):
        for index, value in enumerate(serialized_obj["value"]):
            new_path = f"{parent_path}[{index}]"
            paths.append(new_path)
            if serialized_dict_is_nested_object(value):
                paths.extend(get_data_paths_from_serialized_object(value, new_path))

    elif serialized_dict_is_nested_object(serialized_obj):
        for key, value in cast(
            dict[str, SerializedObject], serialized_obj["value"]
        ).items():
            # Serialized dictionaries need to have a different new_path than nested
            # classes
            if serialized_obj["type"] == "dict":
                processed_key = key
                if isinstance(key, str):
                    processed_key = f'"{key}"'
                new_path = f"{parent_path}[{processed_key}]"
            else:
                new_path = f"{parent_path}.{key}" if parent_path != "" else key

            paths.append(new_path)
            if serialized_dict_is_nested_object(value):
                paths.extend(get_data_paths_from_serialized_object(value, new_path))

    return paths


def generate_serialized_data_paths(
    data: dict[str, SerializedObject],
) -> list[str]:
    """
    Recursively extracts full access paths from a serialized DataService class instance.

    Args:
        data:
            The value of the "value" key of a serialized DataService class instance.

    Returns:
        A list of strings, each representing a full access path in the serialized
            object.
    """

    paths: list[str] = []

    for key, value in data.items():
        paths.append(key)

        if serialized_dict_is_nested_object(value):
            paths.extend(get_data_paths_from_serialized_object(value, key))
    return paths


def add_prefix_to_full_access_path(
    serialized_obj: SerializedObject, prefix: str
) -> Any:
    """Recursively adds a specified prefix to all full access paths of the serialized
    object.

    Args:
        serialized_obj:
            The serialized object to process.
        prefix:
            The prefix string to prepend to each full access path.

    Returns:
        The modified serialized object with the prefix added to all full access paths.

    Example:
        ```python
        >>> serialized_obj = {
        ...     "full_access_path": "",
        ...     "value": {
        ...         "item": {
        ...             "full_access_path": "some_item_path",
        ...             "value": 1.0
        ...         }
        ...     }
        ... }
        ...
        ... modified_data = add_prefix_to_full_access_path(serialized_obj, 'prefix')
        {"full_access_path": "prefix", "value": {"item": {"full_access_path":
        "prefix.some_item_path", "value": 1.0}}}
        ```
    """

    try:
        if serialized_obj.get("full_access_path", None) is not None:
            serialized_obj["full_access_path"] = (
                prefix + "." + serialized_obj["full_access_path"]
                if serialized_obj["full_access_path"] != ""
                else prefix
            )

        if isinstance(serialized_obj["value"], list):
            for value in serialized_obj["value"]:
                add_prefix_to_full_access_path(cast(SerializedObject, value), prefix)

        elif isinstance(serialized_obj["value"], dict):
            for value in cast(
                dict[str, SerializedObject], serialized_obj["value"]
            ).values():
                add_prefix_to_full_access_path(cast(SerializedObject, value), prefix)
    except (TypeError, KeyError, AttributeError):
        # passed dictionary is not a serialized object
        pass
    return serialized_obj


def serialized_dict_is_nested_object(serialized_dict: SerializedObject) -> bool:
    value = serialized_dict["value"]
    # We are excluding Quantity here as the value corresponding to the "value" key is
    # a dictionary of the form {"magnitude": ..., "unit": ...}
    return serialized_dict["type"] != "Quantity" and (isinstance(value, dict | list))


__all__ = ["Serializer", "dump"]



import logging
from typing import TYPE_CHECKING, Any, cast

from pydase.utils.serialization.serializer import (
    SerializedObject,
    get_nested_dict_by_path,
    set_nested_value_by_path,
)

if TYPE_CHECKING:
    from pydase import DataService

logger = logging.getLogger(__name__)


class DataServiceCache:
    """Maintains a serialized cache of the current state of a DataService instance.

    This class is responsible for storing and updating a representation of the service's
    public attributes and properties. It is primarily used by the StateManager and the
    web server to serve consistent state to clients without accessing the DataService
    attributes directly.

    The cache is initialized once upon construction by serializing the full state of
    the service. After that, it can be incrementally updated using attribute paths and
    values as notified by the
    [`DataServiceObserver`][pydase.data_service.data_service_observer.DataServiceObserver].

    Args:
        service: The DataService instance whose state should be cached.
    """

    def __init__(self, service: "DataService") -> None:
        self._cache: SerializedObject
        self.service = service
        self._initialize_cache()

    @property
    def cache(self) -> SerializedObject:
        return self._cache

    def _initialize_cache(self) -> None:
        """Initializes the cache and sets up the callback."""
        logger.debug("Initializing cache.")
        self._cache = self.service.serialize()

    def update_cache(self, full_access_path: str, value: Any) -> None:
        set_nested_value_by_path(
            cast(dict[str, SerializedObject], self._cache["value"]),
            full_access_path,
            value,
        )

    def get_value_dict_from_cache(self, full_access_path: str) -> SerializedObject:
        return get_nested_dict_by_path(
            cast(dict[str, SerializedObject], self._cache["value"]),
            full_access_path,
        )



import asyncio
import logging
import logging.config
import sys
from collections.abc import Callable
from copy import copy
from typing import ClassVar, Literal, TextIO

import click
import socketio  # type: ignore[import-untyped]

import pydase.config

logger = logging.getLogger(__name__)

if pydase.config.OperationMode().environment == "development":
    LOG_LEVEL = logging.DEBUG
else:
    LOG_LEVEL = logging.INFO

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "()": "pydase.utils.logging.DefaultFormatter",
            "fmt": "%(asctime)s.%(msecs)03d | %(levelprefix)s | "
            "%(name)s:%(funcName)s:%(lineno)d - %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        },
    },
    "filters": {
        "only_pydase_server": {
            "()": "pydase.utils.logging.NameFilter",
            "match": "pydase.server",
        },
        "exclude_pydase_server": {
            "()": "pydase.utils.logging.NameFilter",
            "match": "pydase.server",
            "invert": True,
        },
    },
    "handlers": {
        "stdout_handler": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
            "filters": ["only_pydase_server"],
        },
        "stderr_handler": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
            "filters": ["exclude_pydase_server"],
        },
    },
    "loggers": {
        "pydase": {
            "handlers": ["stdout_handler", "stderr_handler"],
            "level": LOG_LEVEL,
            "propagate": False,
        },
        "aiohttp_middlewares": {
            "handlers": ["stderr_handler"],
            "level": logging.WARNING,
            "propagate": False,
        },
        "aiohttp": {
            "handlers": ["stderr_handler"],
            "level": logging.INFO,
            "propagate": False,
        },
    },
}


class NameFilter(logging.Filter):
    """
    Logging filter that allows filtering logs based on the logger name.
    Can either include or exclude a specific logger.
    """

    def __init__(self, match: str, invert: bool = False):
        super().__init__()
        self.match = match
        self.invert = invert

    def filter(self, record: logging.LogRecord) -> bool:
        if self.invert:
            return not record.name.startswith(self.match)
        return record.name.startswith(self.match)


class DefaultFormatter(logging.Formatter):
    """
    A custom log formatter class that:

    * Outputs the LOG_LEVEL with an appropriate color.
    * If a log call includes an `extras={"color_message": ...}` it will be used
      for formatting the output, instead of the plain text message.
    """

    level_name_colors: ClassVar[dict[int, Callable[..., str]]] = {
        logging.DEBUG: lambda level_name: click.style(str(level_name), fg="cyan"),
        logging.INFO: lambda level_name: click.style(str(level_name), fg="green"),
        logging.WARNING: lambda level_name: click.style(str(level_name), fg="yellow"),
        logging.ERROR: lambda level_name: click.style(str(level_name), fg="red"),
        logging.CRITICAL: lambda level_name: click.style(
            str(level_name), fg="bright_red"
        ),
    }

    def __init__(
        self,
        fmt: str | None = None,
        datefmt: str | None = None,
        style: Literal["%", "{", "$"] = "%",
        use_colors: bool | None = None,
    ):
        if use_colors in (True, False):
            self.use_colors = use_colors
        else:
            self.use_colors = sys.stdout.isatty()
        super().__init__(fmt=fmt, datefmt=datefmt, style=style)

    def color_level_name(self, level_name: str, level_no: int) -> str:
        def default(level_name: str) -> str:
            return str(level_name)

        func = self.level_name_colors.get(level_no, default)
        return func(level_name)

    def formatMessage(self, record: logging.LogRecord) -> str:  # noqa: N802
        recordcopy = copy(record)
        levelname = recordcopy.levelname
        seperator = " " * (8 - len(recordcopy.levelname))
        if self.use_colors:
            levelname = self.color_level_name(levelname, recordcopy.levelno)
            if "color_message" in recordcopy.__dict__:
                recordcopy.msg = recordcopy.__dict__["color_message"]
                recordcopy.__dict__["message"] = recordcopy.getMessage()
        recordcopy.__dict__["levelprefix"] = levelname + seperator
        return logging.Formatter.formatMessage(self, recordcopy)

    def should_use_colors(self) -> bool:
        return sys.stderr.isatty()


class SocketIOHandler(logging.Handler):
    """
    Custom logging handler that emits ERROR and CRITICAL log records to a Socket.IO
    server, allowing for real-time logging in applications that use Socket.IO for
    communication.
    """

    def __init__(self, sio: socketio.AsyncServer) -> None:
        super().__init__(logging.ERROR)
        self._sio = sio

    def format(self, record: logging.LogRecord) -> str:
        msg = record.getMessage()
        return f"{record.name}:{record.funcName}:{record.lineno} - {msg}"

    def emit(self, record: logging.LogRecord) -> None:
        log_entry = self.format(record)

        loop = asyncio.get_event_loop()
        loop.create_task(
            self._sio.emit(
                "log",
                {
                    "levelname": record.levelname,
                    "message": log_entry,
                },
            )
        )


def setup_logging() -> None:
    """
    Configures the logging settings for the application.

    This function sets up logging with specific formatting and colorization of log
    messages. The log level is determined based on the application's operation mode. By
    default, in a development environment, the log level is set to DEBUG, whereas in
    other environments, it is set to INFO.
    """

    logger.debug("Configuring pydase logging.")

    logging.config.dictConfig(LOGGING_CONFIG)


def configure_logging_with_pydase_formatter(
    name: str | None = None, level: int = logging.INFO, stream: TextIO | None = None
) -> None:
    """Configure a logger with the pydase `DefaultFormatter`.

    This sets up a `StreamHandler` with the custom `DefaultFormatter`, which includes
    timestamp, log level with color (if supported), logger name, function, and line
    number. It can be used to configure the root logger or any named logger.

    Args:
        name: The name of the logger to configure. If None, the root logger is used.
        level: The logging level to set on the logger (e.g., logging.DEBUG,
            logging.INFO). Defaults to logging.INFO.
        stream: The output stream for the log messages (e.g., sys.stdout or sys.stderr).
            If None, defaults to sys.stderr.

    Example:
        Configure logging in your service:

        ```python
        import sys
        from pydase.utils.logging import configure_logging_with_pydase_formatter

        configure_logging_with_pydase_formatter(
            name="my_service",      # Use the package/module name or None for the root logger
            level=logging.DEBUG,    # Set the desired logging level (defaults to INFO)
            stream=sys.stdout       # Set the output stream (stderr by default)
        )
        ```

    Notes:
        - This function adds a new handler each time it's called.
          Use carefully to avoid duplicate logs.
        - Colors are enabled if the stream supports TTY (e.g., in terminal).
    """  # noqa: E501

    logger = logging.getLogger(name=name)
    handler = logging.StreamHandler(stream=stream)
    formatter = DefaultFormatter(
        fmt="%(asctime)s.%(msecs)03d | %(levelprefix)s | "
        "%(name)s:%(funcName)s:%(lineno)d - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(level)



from typing import TYPE_CHECKING, Any

import pydase.utils.serialization.deserializer
import pydase.utils.serialization.serializer
from pydase.data_service.state_manager import StateManager
from pydase.server.web_server.sio_setup import TriggerMethodDict, UpdateDict
from pydase.utils.helpers import get_object_attr_from_path
from pydase.utils.serialization.types import SerializedObject

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

loads = pydase.utils.serialization.deserializer.loads
Serializer = pydase.utils.serialization.serializer.Serializer


def update_value(state_manager: StateManager, data: UpdateDict) -> None:
    path = data["access_path"]

    state_manager.set_service_attribute_value_by_path(
        path=path, serialized_value=data["value"]
    )


def get_value(state_manager: StateManager, access_path: str) -> SerializedObject:
    return Serializer.serialize_object(
        get_object_attr_from_path(state_manager.service, access_path),
        access_path=access_path,
    )


def trigger_method(state_manager: StateManager, data: TriggerMethodDict) -> Any:
    method = get_object_attr_from_path(state_manager.service, data["access_path"])

    serialized_args = data.get("args", None)
    args = loads(serialized_args) if serialized_args else []

    serialized_kwargs = data.get("kwargs", None)
    kwargs: dict[str, Any] = loads(serialized_kwargs) if serialized_kwargs else {}

    return Serializer.serialize_object(method(*args, **kwargs))


async def trigger_async_method(
    state_manager: StateManager, data: TriggerMethodDict
) -> Any:
    method: Callable[..., Awaitable[Any]] = get_object_attr_from_path(
        state_manager.service, data["access_path"]
    )

    serialized_args = data.get("args", None)
    args = loads(serialized_args) if serialized_args else []

    serialized_kwargs = data.get("kwargs", None)
    kwargs: dict[str, Any] = loads(serialized_kwargs) if serialized_kwargs else {}

    return Serializer.serialize_object(await method(*args, **kwargs))



import inspect
import logging
from functools import partial
from typing import TYPE_CHECKING

import aiohttp.web
import aiohttp_middlewares.error
import click

from pydase.data_service.state_manager import StateManager
from pydase.server.web_server.api.v1.endpoints import (
    get_value,
    trigger_async_method,
    trigger_method,
    update_value,
)
from pydase.utils.helpers import get_object_attr_from_path
from pydase.utils.serialization.serializer import dump

if TYPE_CHECKING:
    from pydase.server.web_server.sio_setup import TriggerMethodDict, UpdateDict

logger = logging.getLogger(__name__)

STATUS_OK = 200
STATUS_FAILED = 400


async def _get_value(
    request: aiohttp.web.Request, state_manager: StateManager
) -> aiohttp.web.Response:
    log_id = get_log_id(request)

    access_path = request.rel_url.query["access_path"]

    logger.info("Client [%s] is getting the value of '%s'", log_id, access_path)

    status = STATUS_OK
    try:
        result = get_value(state_manager, access_path)
    except Exception as e:
        logger.exception(e)
        result = dump(e)
        status = STATUS_FAILED
    return aiohttp.web.json_response(result, status=status)


async def _update_value(
    request: aiohttp.web.Request, state_manager: StateManager
) -> aiohttp.web.Response:
    log_id = get_log_id(request)

    data: UpdateDict = await request.json()

    logger.info(
        "Client [%s] is updating the value of '%s'", log_id, data["access_path"]
    )

    try:
        update_value(state_manager, data)

        return aiohttp.web.json_response()
    except Exception as e:
        logger.exception(e)
        return aiohttp.web.json_response(dump(e), status=STATUS_FAILED)


async def _trigger_method(
    request: aiohttp.web.Request, state_manager: StateManager
) -> aiohttp.web.Response:
    log_id = get_log_id(request)

    data: TriggerMethodDict = await request.json()

    access_path = data["access_path"]

    logger.info("Client [%s] is triggering the method '%s'", log_id, access_path)

    method = get_object_attr_from_path(state_manager.service, access_path)

    try:
        if inspect.iscoroutinefunction(method):
            method_return = await trigger_async_method(
                state_manager=state_manager, data=data
            )
        else:
            method_return = trigger_method(state_manager=state_manager, data=data)

        return aiohttp.web.json_response(method_return)

    except Exception as e:
        logger.exception(e)
        return aiohttp.web.json_response(dump(e), status=STATUS_FAILED)


def get_log_id(request: aiohttp.web.Request) -> str:
    client_id_header = request.headers.get("x-client-id", None)
    remote_username_header = request.headers.get("remote-user", None)

    if remote_username_header is not None:
        log_id = f"user={click.style(remote_username_header, fg='cyan')}"
    elif client_id_header is not None:
        log_id = f"id={click.style(client_id_header, fg='cyan')}"
    else:
        log_id = f"id={click.style(None, fg='cyan')}"

    return log_id


def create_api_application(state_manager: StateManager) -> aiohttp.web.Application:
    api_application = aiohttp.web.Application(
        middlewares=(aiohttp_middlewares.error.error_middleware(),)
    )

    api_application.router.add_get(
        "/get_value", partial(_get_value, state_manager=state_manager)
    )
    api_application.router.add_put(
        "/update_value", partial(_update_value, state_manager=state_manager)
    )
    api_application.router.add_put(
        "/trigger_method", partial(_trigger_method, state_manager=state_manager)
    )

    return api_application



from enum import Enum


class ColouredEnum(Enum):
    """
    Represents a UI element that can display colour-coded text based on its value.

    This class extends the standard Enum but requires its values to be valid CSS
    colour codes. Supported colour formats include:

    - Hexadecimal colours
    - Hexadecimal colours with transparency
    - RGB colours
    - RGBA colours
    - HSL colours
    - HSLA colours
    - Predefined/Cross-browser colour names

    Refer to the this website for more details on colour formats:
    (https://www.w3schools.com/cssref/css_colours_legal.php)

    The behavior of this component in the UI depends on how it's defined in the data
    service:

    - As property with a setter or as attribute: Renders as a dropdown menu, allowing
    users to select and change its value from the frontend.
    - As property without a setter: Displays as a coloured box with the key of the
    `ColouredEnum` as text inside, serving as a visual indicator without user
    interaction.

    Example:
        ```python
        import pydase.components as pyc
        import pydase

        class MyStatus(pyc.ColouredEnum):
            PENDING = "#FFA500"  # Orange
            RUNNING = "#0000FF80"  # Transparent Blue
            PAUSED = "rgb(169, 169, 169)"  # Dark Gray
            RETRYING = "rgba(255, 255, 0, 0.3)"  # Transparent Yellow
            COMPLETED = "hsl(120, 100%, 50%)"  # Green
            FAILED = "hsla(0, 100%, 50%, 0.7)"  # Transparent Red
            CANCELLED = "SlateGray"  # Slate Gray

        class StatusExample(pydase.DataService):
            _status = MyStatus.RUNNING

            @property
            def status(self) -> MyStatus:
                return self._status

            @status.setter
            def status(self, value: MyStatus) -> None:
                # Custom logic here...
                self._status = value

        # Example usage:
        my_service = StatusExample()
        my_service.status = MyStatus.FAILED
        ```

    Note:
        Each enumeration name and value must be unique. This means that you should use
        different colour formats when you want to use a colour multiple times.
    """



import logging
from typing import Any

from pydase.data_service.data_service import DataService

logger = logging.getLogger(__name__)


class NumberSlider(DataService):
    """
    This class models a UI slider for a data service, allowing for adjustments of a
    parameter within a specified range and increments.

    Args:
        value:
            The initial value of the slider. Defaults to 0.0.
        min_:
            The minimum value of the slider. Defaults to 0.0.
        max_:
            The maximum value of the slider. Defaults to 100.0.
        step_size:
            The increment/decrement step size of the slider. Defaults to 1.0.

    Example:
        ```python
        class MySlider(pydase.components.NumberSlider):
            def __init__(
                self,
                value: float = 0.0,
                min_: float = 0.0,
                max_: float = 100.0,
                step_size: float = 1.0,
            ) -> None:
                super().__init__(value, min_, max_, step_size)

            @property
            def min(self) -> float:
                return self._min

            @min.setter
            def min(self, value: float) -> None:
                self._min = value

            @property
            def max(self) -> float:
                return self._max

            @max.setter
            def max(self, value: float) -> None:
                self._max = value

            @property
            def step_size(self) -> float:
                return self._step_size

            @step_size.setter
            def step_size(self, value: float) -> None:
                self._step_size = value

            @property
            def value(self) -> float:
                return self._value

            @value.setter
            def value(self, value: float) -> None:
                if value < self._min or value > self._max:
                    raise ValueError(
                        "Value is either below allowed min or above max value."
                    )

                self._value = value

        class MyService(pydase.DataService):
            def __init__(self) -> None:
                self.voltage = MyService()

        # Modifying or accessing the voltage value:
        my_service = MyService()
        my_service.voltage.value = 5
        print(my_service.voltage.value)  # Output: 5
        ```
    """

    def __init__(
        self,
        value: Any = 0.0,
        min_: Any = 0.0,
        max_: Any = 100.0,
        step_size: Any = 1.0,
    ) -> None:
        super().__init__()
        self._step_size = step_size
        self._value = value
        self._min = min_
        self._max = max_

    @property
    def min(self) -> Any:
        """The min property."""
        return self._min

    @property
    def max(self) -> Any:
        """The min property."""
        return self._max

    @property
    def step_size(self) -> Any:
        """The min property."""
        return self._step_size

    @property
    def value(self) -> Any:
        """The value property."""
        return self._value

    @value.setter
    def value(self, value: Any) -> None:
        self._value = value



import logging

import aiohttp.web
import aiohttp_middlewares.error

import pydase.server.web_server.api.v1.application
from pydase.data_service.state_manager import StateManager

logger = logging.getLogger(__name__)


def create_api_application(state_manager: StateManager) -> aiohttp.web.Application:
    api_application = aiohttp.web.Application(
        middlewares=(aiohttp_middlewares.error.error_middleware(),)
    )

    api_application.add_subapp(
        "/v1/",
        pydase.server.web_server.api.v1.application.create_api_application(
            state_manager
        ),
    )

    return api_application
